"""Initializer of Auto Instrumentation of Pydantic AI Functions"""

from typing import Collection
import importlib.metadata
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from wrapt import wrap_function_wrapper

from openlit.instrumentation.pydantic_ai.pydantic_ai import (
    agent_create, agent_run, async_agent_run
)

_instruments = ('pydantic-ai >= 0.2.17',)

class PydanticAIInstrumentor(BaseInstrumentor):
    """
    An instrumentor for Pydantic AI's client library.
    """

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        application_name = kwargs.get('application_name', 'default_application')
        environment = kwargs.get('environment', 'default_environment')
        tracer = kwargs.get('tracer')
        metrics = kwargs.get('metrics_dict')
        pricing_info = kwargs.get('pricing_info', {})
        capture_message_content = kwargs.get('capture_message_content', False)
        disable_metrics = kwargs.get('disable_metrics')
        version = importlib.metadata.version('pydantic-ai')

        wrap_function_wrapper(
            'pydantic_ai.agent',
            'Agent.__init__',
            agent_create(version, environment, application_name,
                  tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )

        wrap_function_wrapper(
            'pydantic_ai.agent',
            'Agent.run_sync',
            agent_run(version, environment, application_name,
                  tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )

        wrap_function_wrapper(
            'pydantic_ai.agent',
            'Agent.run',
            async_agent_run(version, environment, application_name,
                  tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )

    def _uninstrument(self, **kwargs):
        # Proper uninstrumentation logic to revert patched methods
        pass



"""
Pydantic AI OpenTelemetry instrumentation utility functions
"""
import logging
from opentelemetry.sdk.resources import SERVICE_NAME, TELEMETRY_SDK_NAME, DEPLOYMENT_ENVIRONMENT
from opentelemetry.trace import Status, StatusCode, SpanKind
from openlit.__helpers import (
    handle_exception
)
from openlit.semcov import SemanticConvention

# Initialize logger for logging potential issues and operations
logger = logging.getLogger(__name__)

def set_span_attributes(span, version, operation_name, environment,
        application_name, server_address, server_port, request_model, agent_name):
    """
    Set common OpenTelemetry span attributes for Pydantic AI operations.
    """

    # Set Span attributes (OTel Semconv)
    span.set_attribute(TELEMETRY_SDK_NAME, 'openlit')
    span.set_attribute(SemanticConvention.GEN_AI_OPERATION, operation_name)
    span.set_attribute(SemanticConvention.GEN_AI_SYSTEM, SemanticConvention.GEN_AI_SYSTEM_PYDANTIC_AI)
    span.set_attribute(SemanticConvention.GEN_AI_AGENT_NAME, agent_name)
    span.set_attribute(SemanticConvention.SERVER_ADDRESS, server_address)
    span.set_attribute(SemanticConvention.SERVER_PORT, server_port)
    span.set_attribute(SemanticConvention.GEN_AI_REQUEST_MODEL, request_model)

    # Set Span attributes (Extras)
    span.set_attribute(DEPLOYMENT_ENVIRONMENT, environment)
    span.set_attribute(SERVICE_NAME, application_name)
    span.set_attribute(SemanticConvention.GEN_AI_SDK_VERSION, version)

def common_agent_run(wrapped, instance, args, kwargs, tracer, version, environment, application_name,
                         capture_message_content, response):
    """
    Handle telemetry for Pydantic AI agent run operations.
    """

    server_address, server_port = instance.model.base_url, 443
    agent_name = instance.name or "pydantic_agent"
    request_model = str(instance.model.model_name)
    span_name = f'{SemanticConvention.GEN_AI_OPERATION_TYPE_EXECUTE_AGENT_TASK} {agent_name}'

    with tracer.start_as_current_span(span_name, kind=SpanKind.CLIENT) as span:
        try:
            set_span_attributes(span, version, SemanticConvention.GEN_AI_OPERATION_TYPE_EXECUTE_AGENT_TASK,
                                environment, application_name, server_address, server_port, request_model, agent_name)
            span.set_attribute(SemanticConvention.GEN_AI_AGENT_DESCRIPTION, str(instance._system_prompts))
            span.set_attribute(SemanticConvention.GEN_AI_RESPONSE_MODEL, request_model)

            if capture_message_content:
                span.add_event(
                    name=SemanticConvention.GEN_AI_CONTENT_COMPLETION_EVENT,
                    attributes={
                        SemanticConvention.GEN_AI_CONTENT_COMPLETION: response.output,
                    },
                )

            span.set_status(Status(StatusCode.OK))

            return response

        except Exception as e:
            handle_exception(span, e)
            logger.error('Error in trace creation: %s', e)
            return response

def common_agent_create(wrapped, instance, args, kwargs, tracer, version, environment, application_name,
                       capture_message_content, response):
    """
    Handle telemetry for Pydantic AI agent creation operations.
    """

    server_address, server_port = '127.0.0.1', 80
    agent_name = kwargs.get("name", "pydantic_agent")
    span_name = f'{SemanticConvention.GEN_AI_OPERATION_TYPE_CREATE_AGENT} {agent_name}'

    with tracer.start_as_current_span(span_name, kind=SpanKind.CLIENT) as span:
        try:
            request_model = args[0] or kwargs.get("model", "google-gla:gemini-1.5-flash")
            set_span_attributes(span, version, SemanticConvention.GEN_AI_OPERATION_TYPE_CREATE_AGENT,
                environment, application_name, server_address, server_port, request_model, agent_name)
            span.set_attribute(SemanticConvention.GEN_AI_AGENT_DESCRIPTION, str(kwargs.get("system_prompt", "")))
            span.set_attribute(SemanticConvention.GEN_AI_RESPONSE_MODEL, request_model)

            span.set_status(Status(StatusCode.OK))

            return response

        except Exception as e:
            handle_exception(span, e)
            logger.error('Error in trace creation: %s', e)
            return response



"""
Astra OpenTelemetry instrumentation utility functions
"""

import time
import logging
from opentelemetry.trace import Status, StatusCode
from opentelemetry.sdk.resources import SERVICE_NAME, TELEMETRY_SDK_NAME, DEPLOYMENT_ENVIRONMENT
from openlit.__helpers import handle_exception
from openlit.semcov import SemanticConvention

# Initialize logger for logging potential issues and operations
logger = logging.getLogger(__name__)

def object_count(obj):
    """Counts Length of object if it exists, Else returns None."""
    return len(obj) if isinstance(obj, list) else 1

DB_OPERATION_MAP = {
    'astra.create_collection': SemanticConvention.DB_OPERATION_CREATE_COLLECTION,
    'astra.drop_collection': SemanticConvention.DB_OPERATION_DELETE_COLLECTION,
    'astra.insert': SemanticConvention.DB_OPERATION_INSERT,
    'astra.update': SemanticConvention.DB_OPERATION_UPDATE,
    'astra.find': SemanticConvention.DB_OPERATION_SELECT,
    'astra.find_one_and_update': SemanticConvention.DB_OPERATION_REPLACE,
    'astra.replace_one': SemanticConvention.DB_OPERATION_REPLACE,
    'astra.delete': SemanticConvention.DB_OPERATION_DELETE,
    'astra.find_one_and_delete': SemanticConvention.DB_OPERATION_FIND_AND_DELETE
}

def process_db_operations(response, span, start_time, gen_ai_endpoint,
        version, environment, application_name,
        capture_message_content, metrics, disable_metrics, server_address,
        server_port, collection_name, db_operation, kwargs, args):
    """
    Process DB operation and generate Telemetry
    """

    end_time = time.time()

    try:
        span.set_attribute(TELEMETRY_SDK_NAME, 'openlit')
        span.set_attribute(SemanticConvention.GEN_AI_OPERATION, SemanticConvention.GEN_AI_OPERATION_TYPE_VECTORDB)
        span.set_attribute(SemanticConvention.DB_SYSTEM_NAME, SemanticConvention.DB_SYSTEM_ASTRA)
        span.set_attribute(SemanticConvention.DB_CLIENT_OPERATION_DURATION, end_time - start_time)
        span.set_attribute(SemanticConvention.SERVER_ADDRESS, server_address)
        span.set_attribute(SemanticConvention.SERVER_PORT, server_port)
        span.set_attribute(DEPLOYMENT_ENVIRONMENT, environment)
        span.set_attribute(SERVICE_NAME, application_name)
        span.set_attribute(SemanticConvention.DB_OPERATION_NAME, db_operation)
        span.set_attribute(SemanticConvention.DB_COLLECTION_NAME, collection_name)
        span.set_attribute(SemanticConvention.DB_SDK_VERSION, version)

        if db_operation == SemanticConvention.DB_OPERATION_CREATE_COLLECTION:
            span.set_attribute(SemanticConvention.DB_NAMESPACE, response.keyspace)
            span.set_attribute(SemanticConvention.DB_COLLECTION_NAME, response.name)
            span.set_attribute(SemanticConvention.DB_INDEX_DIMENSION, kwargs.get('dimension', ''))
            span.set_attribute(SemanticConvention.DB_INDEX_METRIC, str(kwargs.get('metric', '')))

        if db_operation == SemanticConvention.DB_OPERATION_INSERT:
            span.set_attribute(SemanticConvention.DB_DOCUMENTS_COUNT, object_count(args[0]))
            span.set_attribute(SemanticConvention.DB_QUERY_TEXT, str(args[0] or kwargs.get('documents', {})))

        elif db_operation == SemanticConvention.DB_OPERATION_UPDATE:
            span.set_attribute(SemanticConvention.DB_RESPONSE_RETURNED_ROWS, response.update_info.get('nModified', 0))
            span.set_attribute(SemanticConvention.DB_QUERY_TEXT, str(args[1] or kwargs.get('update', {})))

        elif db_operation == SemanticConvention.DB_OPERATION_DELETE:
            span.set_attribute(SemanticConvention.DB_RESPONSE_RETURNED_ROWS, response.deleted_count)
            span.set_attribute(SemanticConvention.DB_QUERY_TEXT, str(args[0] or kwargs.get('filter', {})))

        elif db_operation in [
            SemanticConvention.DB_OPERATION_SELECT,
            SemanticConvention.DB_OPERATION_FIND_AND_DELETE,
            SemanticConvention.DB_OPERATION_REPLACE
        ]:
            span.set_attribute(SemanticConvention.DB_QUERY_TEXT, str(args or kwargs.get('filter', {})))

        span.set_status(Status(StatusCode.OK))

        if not disable_metrics:
            attributes = {
                TELEMETRY_SDK_NAME: 'openlit',
                SERVICE_NAME: application_name,
                SemanticConvention.DB_SYSTEM_NAME: SemanticConvention.DB_SYSTEM_ASTRA,
                DEPLOYMENT_ENVIRONMENT: environment,
                SemanticConvention.GEN_AI_OPERATION: SemanticConvention.GEN_AI_OPERATION_TYPE_VECTORDB,
                SemanticConvention.DB_OPERATION_NAME: db_operation
            }

            metrics['db_requests'].add(1, attributes)
            metrics['db_client_operation_duration'].record(end_time - start_time, attributes)

        # Return original response
        return response

    except Exception as e:
        handle_exception(span, e)
        logger.error('Error in trace creation: %s', e)

        # Return original response
        return response



"""
LiteLLM OpenTelemetry instrumentation utility functions
"""
import time

from opentelemetry.trace import Status, StatusCode

from openlit.__helpers import (
    calculate_ttft,
    response_as_dict,
    calculate_tbt,
    get_chat_model_cost,
    get_embed_model_cost,
    common_span_attributes,
    record_completion_metrics,
    record_embedding_metrics,
)
from openlit.semcov import SemanticConvention

def format_content(messages):
    """
    Process a list of messages to extract content.
    """

    formatted_messages = []
    for message in messages:
        role = message['role']
        content = message['content']

        if isinstance(content, list):
            content_str = ", ".join(
                f'{item["type"]}: {item["text"] if "text" in item else item["image_url"]}'
                if "type" in item else f'text: {item["text"]}'
                for item in content
            )
            formatted_messages.append(f'{role}: {content_str}')
        else:
            formatted_messages.append(f'{role}: {content}')

    return '\n'.join(formatted_messages)

def process_chunk(scope, chunk):
    """
    Process a chunk of response data and update state.
    """

    end_time = time.time()
    # Record the timestamp for the current chunk
    scope._timestamps.append(end_time)

    if len(scope._timestamps) == 1:
        # Calculate time to first chunk
        scope._ttft = calculate_ttft(scope._timestamps, scope._start_time)

    chunked = response_as_dict(chunk)

    # Collect message IDs and aggregated response from events
    if (len(chunked.get('choices', [])) > 0 and ('delta' in chunked.get('choices')[0] and
        'content' in chunked.get('choices')[0].get('delta', {}))):

        content = chunked.get('choices')[0].get('delta').get('content')
        if content:
            scope._llmresponse += content

        # Handle tool calls in streaming - optimized
        delta_tools = chunked.get('choices', [{}])[0].get('delta', {}).get('tool_calls')
        if delta_tools:
            scope._tools = scope._tools or []

            for tool in delta_tools:
                idx = tool.get('index', 0)

                # Extend list if needed
                scope._tools.extend([{}] * (idx + 1 - len(scope._tools)))

                if tool.get('id'):  # New tool (id exists)
                    func = tool.get('function', {})
                    scope._tools[idx] = {
                        'id': tool['id'],
                        'function': {'name': func.get('name', ''), 'arguments': func.get('arguments', '')},
                        'type': tool.get('type', 'function')
                    }
                elif scope._tools[idx] and 'function' in tool:  # Append args (id is None)
                    scope._tools[idx]['function']['arguments'] += tool['function'].get('arguments', '')

    if chunked.get('usage'):
        scope._input_tokens = chunked.get('usage').get('prompt_tokens', 0)
        scope._output_tokens = chunked.get('usage').get('completion_tokens', 0)
        scope._response_id = chunked.get('id')
        scope._response_model = chunked.get('model')
        scope._finish_reason = chunked.get('choices', [{}])[0].get('finish_reason')
        scope._response_service_tier = str(chunked.get('system_fingerprint', ''))
        scope._end_time = time.time()

def common_chat_logic(scope, pricing_info, environment, application_name, metrics,
    capture_message_content, disable_metrics, version, is_stream):
    """
    Process chat request and generate Telemetry
    """

    if len(scope._timestamps) > 1:
        scope._tbt = calculate_tbt(scope._timestamps)

    prompt = format_content(scope._kwargs.get('messages', []))
    request_model = scope._kwargs.get('model', 'openai/gpt-4o')

    cost = get_chat_model_cost(request_model, pricing_info, scope._input_tokens, scope._output_tokens)

    # Common Span Attributes
    common_span_attributes(scope,
        SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT, SemanticConvention.GEN_AI_SYSTEM_LITELLM,
        scope._server_address, scope._server_port, request_model, scope._response_model,
        environment, application_name, is_stream, scope._tbt, scope._ttft, version)

    # Span Attributes for Request parameters
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_SEED, scope._kwargs.get('seed', ''))
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_FREQUENCY_PENALTY, scope._kwargs.get('frequency_penalty', 0.0))
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_MAX_TOKENS, scope._kwargs.get('max_tokens', -1))
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_PRESENCE_PENALTY, scope._kwargs.get('presence_penalty', 0.0))
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_STOP_SEQUENCES, scope._kwargs.get('stop', []))
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_TEMPERATURE, scope._kwargs.get('temperature', 1.0))
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_TOP_P, scope._kwargs.get('top_p', 1.0))
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_USER, scope._kwargs.get('user', ''))
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_SERVICE_TIER, scope._kwargs.get('service_tier', 'auto'))

    # Span Attributes for Response parameters
    scope._span.set_attribute(SemanticConvention.GEN_AI_RESPONSE_ID, scope._response_id)
    scope._span.set_attribute(SemanticConvention.GEN_AI_RESPONSE_FINISH_REASON, [scope._finish_reason])
    scope._span.set_attribute(SemanticConvention.GEN_AI_RESPONSE_SERVICE_TIER, scope._response_service_tier)
    scope._span.set_attribute(SemanticConvention.GEN_AI_RESPONSE_SYSTEM_FINGERPRINT, scope._response_service_tier)
    scope._span.set_attribute(SemanticConvention.GEN_AI_OUTPUT_TYPE, "text" if isinstance(scope._llmresponse, str) else "json")

    # Span Attributes for Cost and Tokens
    scope._span.set_attribute(SemanticConvention.GEN_AI_USAGE_INPUT_TOKENS, scope._input_tokens)
    scope._span.set_attribute(SemanticConvention.GEN_AI_USAGE_OUTPUT_TOKENS, scope._output_tokens)
    scope._span.set_attribute(SemanticConvention.GEN_AI_CLIENT_TOKEN_USAGE, scope._input_tokens + scope._output_tokens)
    scope._span.set_attribute(SemanticConvention.GEN_AI_USAGE_COST, cost)

    # Span Attributes for Tools - optimized
    if scope._tools:
        tools = scope._tools if isinstance(scope._tools, list) else [scope._tools]

        names, ids, args = zip(*[
            (t.get("function", {}).get("name", ""),
             str(t.get("id", "")),
             str(t.get("function", {}).get("arguments", "")))
            for t in tools if isinstance(t, dict) and t
        ]) if tools else ([], [], [])

        scope._span.set_attribute(SemanticConvention.GEN_AI_TOOL_NAME, ", ".join(filter(None, names)))
        scope._span.set_attribute(SemanticConvention.GEN_AI_TOOL_CALL_ID, ", ".join(filter(None, ids)))
        scope._span.set_attribute(SemanticConvention.GEN_AI_TOOL_ARGS, ", ".join(filter(None, args)))

    # Span Attributes for Content
    if capture_message_content:
        scope._span.set_attribute(SemanticConvention.GEN_AI_CONTENT_PROMPT, prompt)
        scope._span.set_attribute(SemanticConvention.GEN_AI_CONTENT_COMPLETION, scope._llmresponse)

        # To be removed once the change to span_attributes (from span events) is complete
        scope._span.add_event(
            name=SemanticConvention.GEN_AI_CONTENT_PROMPT_EVENT,
            attributes={
                SemanticConvention.GEN_AI_CONTENT_PROMPT: prompt,
            },
        )
        scope._span.add_event(
            name=SemanticConvention.GEN_AI_CONTENT_COMPLETION_EVENT,
            attributes={
                SemanticConvention.GEN_AI_CONTENT_COMPLETION: scope._llmresponse,
            },
        )

    scope._span.set_status(Status(StatusCode.OK))

    # Metrics
    if not disable_metrics:
        record_completion_metrics(metrics, SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT, SemanticConvention.GEN_AI_SYSTEM_LITELLM,
            scope._server_address, scope._server_port, request_model, scope._response_model, environment,
            application_name, scope._start_time, scope._end_time, scope._input_tokens, scope._output_tokens,
            cost, scope._tbt, scope._ttft)

def process_streaming_chat_response(scope, pricing_info, environment, application_name, metrics,
    capture_message_content=False, disable_metrics=False, version=""):
    """
    Process streaming chat request and generate Telemetry
    """

    common_chat_logic(scope, pricing_info, environment, application_name, metrics,
        capture_message_content, disable_metrics, version, is_stream=True)

def process_chat_response(response, request_model, pricing_info, server_port, server_address,
    environment, application_name, metrics, start_time, span, capture_message_content=False,
    disable_metrics=False, version="1.0.0", **kwargs):
    """
    Process chat request and generate Telemetry
    """

    # Create scope object
    scope = type("GenericScope", (), {})()
    response_dict = response_as_dict(response)

    scope._start_time = start_time
    scope._end_time = time.time()
    scope._span = span
    scope._llmresponse = " ".join(
        (choice.get("message", {}).get("content") or "")
        for choice in response_dict.get("choices", [])
    )
    scope._input_tokens = response_dict.get('usage', {}).get('prompt_tokens', 0)
    scope._output_tokens = response_dict.get('usage', {}).get('completion_tokens', 0)
    scope._response_id = response_dict.get('id')
    scope._response_model = response_dict.get('model')
    scope._finish_reason = str(response_dict.get('choices', [])[0].get('finish_reason', ''))
    scope._response_service_tier = str(response_dict.get('system_fingerprint', ''))
    scope._timestamps = []
    scope._ttft, scope._tbt = scope._end_time - scope._start_time, 0
    scope._server_address, scope._server_port = server_address, server_port
    scope._kwargs = kwargs

    # Handle tool calls
    if scope._kwargs.get("tools"):
        scope._tools = response_dict.get("choices", [{}])[0].get("message", {}).get("tool_calls")
    else:
        scope._tools = None

    common_chat_logic(scope, pricing_info, environment, application_name, metrics,
        capture_message_content, disable_metrics, version, is_stream=False)

    return response

def process_embedding_response(response, request_model, pricing_info, server_port, server_address,
    environment, application_name, metrics, start_time, span, capture_message_content=False,
    disable_metrics=False, version="1.0.0", **kwargs):
    """
    Process embedding request and generate Telemetry
    """

    # Create scope object
    scope = type("GenericScope", (), {})()
    response_dict = response_as_dict(response)

    scope._start_time = start_time
    scope._end_time = time.time()
    scope._span = span
    scope._input_tokens = response_dict.get('usage', {}).get('prompt_tokens', 0)
    scope._response_model = response_dict.get('model')
    scope._server_address, scope._server_port = server_address, server_port
    scope._kwargs = kwargs

    # Calculate cost of the operation
    cost = get_embed_model_cost(request_model, pricing_info, scope._input_tokens)

    # Common Span Attributes
    common_span_attributes(scope,
        SemanticConvention.GEN_AI_OPERATION_TYPE_EMBEDDING, SemanticConvention.GEN_AI_SYSTEM_LITELLM,
        scope._server_address, scope._server_port, request_model, scope._response_model,
        environment, application_name, False, 0, scope._end_time - scope._start_time, version)

    # Span Attributes for Request parameters
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_ENCODING_FORMATS, [scope._kwargs.get('encoding_format', 'float')])
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_USER, scope._kwargs.get('user', ''))

    # Span Attributes for Cost and Tokens
    scope._span.set_attribute(SemanticConvention.GEN_AI_USAGE_INPUT_TOKENS, scope._input_tokens)
    scope._span.set_attribute(SemanticConvention.GEN_AI_CLIENT_TOKEN_USAGE, scope._input_tokens)
    scope._span.set_attribute(SemanticConvention.GEN_AI_USAGE_COST, cost)

    # Span Attributes for Content
    if capture_message_content:
        scope._span.set_attribute(SemanticConvention.GEN_AI_CONTENT_PROMPT, str(scope._kwargs.get('input', '')))

        # To be removed once the change to span_attributes (from span events) is complete
        scope._span.add_event(
            name=SemanticConvention.GEN_AI_CONTENT_PROMPT_EVENT,
            attributes={
                SemanticConvention.GEN_AI_CONTENT_PROMPT: str(scope._kwargs.get('input', '')),
            },
        )

    scope._span.set_status(Status(StatusCode.OK))

    # Metrics
    if not disable_metrics:
        record_embedding_metrics(metrics, SemanticConvention.GEN_AI_OPERATION_TYPE_EMBEDDING, SemanticConvention.GEN_AI_SYSTEM_LITELLM,
            scope._server_address, scope._server_port, request_model, scope._response_model, environment,
            application_name, scope._start_time, scope._end_time, scope._input_tokens, cost)

    return response



"""Initializer of Auto Instrumentation of Prem AI Functions"""

from typing import Collection
import importlib.metadata
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from wrapt import wrap_function_wrapper

from openlit.instrumentation.premai.premai import (
    chat, embedding
)

_instruments = ("premai >= 0.3.79",)

class PremAIInstrumentor(BaseInstrumentor):
    """
    An instrumentor for Prem AI's client library.
    """

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        application_name = kwargs.get("application_name", "default")
        environment = kwargs.get("environment", "default")
        tracer = kwargs.get("tracer")
        metrics = kwargs.get("metrics_dict")
        pricing_info = kwargs.get("pricing_info", {})
        capture_message_content = kwargs.get("capture_message_content", False)
        disable_metrics = kwargs.get("disable_metrics")
        version = importlib.metadata.version("premai")

        # sync chat
        wrap_function_wrapper(
            "premai.api",
            "ChatCompletionsModule.create",
            chat(version, environment, application_name,
                  tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )

        # sync embedding
        wrap_function_wrapper(
            "premai.api",
            "EmbeddingsModule.create",
            embedding(version, environment, application_name,
                  tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )

    def _uninstrument(self, **kwargs):
        # Proper uninstrumentation logic to revert patched methods
        pass



"""
Module for monitoring Amazon Bedrock API calls.
"""

import logging
import time
from opentelemetry.trace import SpanKind
from openlit.__helpers import (
    set_server_address_and_port
)
from openlit.instrumentation.bedrock.utils import (
    process_chat_response,
)
from openlit.semcov import SemanticConvention

# Initialize logger for logging potential issues and operations
logger = logging.getLogger(__name__)

def converse(version, environment, application_name, tracer, event_provider,
         pricing_info, capture_message_content, metrics, disable_metrics):
    """
    Generates a telemetry wrapper for GenAI function call
    """

    def wrapper(wrapped, instance, args, kwargs):
        """
        Wraps the GenAI function call.
        """

        def converse_wrapper(original_method, *method_args, **method_kwargs):

            """
            Wraps the GenAI function call.
            """

            server_address, server_port = set_server_address_and_port(instance, 'aws.amazon.com', 443)
            request_model = method_kwargs.get('modelId', 'amazon.titan-text-express-v1')

            span_name = f'{SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT} {request_model}'

            with tracer.start_as_current_span(span_name, kind=SpanKind.CLIENT) as span:
                start_time = time.time()
                response = original_method(*method_args, **method_kwargs)
                llm_config = method_kwargs.get('inferenceConfig', {})
                response = process_chat_response(
                    response=response,
                    request_model=request_model,
                    pricing_info=pricing_info,
                    server_port=server_port,
                    server_address=server_address,
                    environment=environment,
                    application_name=application_name,
                    metrics=metrics,
                    event_provider=event_provider,
                    start_time=start_time,
                    span=span,
                    capture_message_content=capture_message_content,
                    disable_metrics=disable_metrics,
                    version=version,
                    llm_config=llm_config,
                    **method_kwargs
                )

                return response

        # Get the original client instance from the wrapper
        client = wrapped(*args, **kwargs)

        # Replace the original method with the instrumented one
        if kwargs.get('service_name') == 'bedrock-runtime':
            original_invoke_model = client.converse
            client.converse = lambda *args, **kwargs: converse_wrapper(original_invoke_model,
                                                                            *args, **kwargs)

        return client

    return wrapper



"""
Groq OpenTelemetry instrumentation utility functions
"""
import time

from opentelemetry.trace import Status, StatusCode

from openlit.__helpers import (
    calculate_ttft,
    response_as_dict,
    calculate_tbt,
    get_chat_model_cost,
    common_span_attributes,
    record_completion_metrics,
)
from openlit.semcov import SemanticConvention

def format_content(messages):
    """
    Process a list of messages to extract content.
    """

    formatted_messages = []
    for message in messages:
        role = message["role"]
        content = message["content"]

        if isinstance(content, list):
            content_str = ", ".join(
                f'{item["type"]}: {item["text"] if "text" in item else item["image_url"]}'
                if "type" in item else f'text: {item["text"]}'
                for item in content
            )
            formatted_messages.append(f"{role}: {content_str}")
        else:
            formatted_messages.append(f"{role}: {content}")

    return "\n".join(formatted_messages)

def process_chunk(scope, chunk):
    """
    Process a chunk of response data and update state.
    """

    end_time = time.time()
    # Record the timestamp for the current chunk
    scope._timestamps.append(end_time)

    if len(scope._timestamps) == 1:
        # Calculate time to first chunk
        scope._ttft = calculate_ttft(scope._timestamps, scope._start_time)

    chunked = response_as_dict(chunk)

    # Collect message IDs and aggregated response from events
    if (len(chunked.get("choices", [])) > 0 and
        "delta" in chunked.get("choices")[0] and
        "content" in chunked.get("choices")[0].get("delta", {})):

        content = chunked.get("choices")[0].get("delta").get("content")
        if content:
            scope._llmresponse += content

    if chunked.get('x_groq') is not None:
        if chunked.get('x_groq').get('usage') is not None:
            scope._input_tokens = chunked.get('x_groq').get('usage').get('prompt_tokens')
            scope._output_tokens = chunked.get('x_groq').get('usage').get('completion_tokens')
            scope._response_id = chunked.get('x_groq').get('id')
            scope._response_model = chunked.get('x_groq').get('model')
            scope._finish_reason = chunked.get('choices', [{}])[0].get('finish_reason')
            scope._system_fingerprint = chunked.get('x_groq').get('system_fingerprint')
            scope._end_time = time.time()


def common_chat_logic(scope, pricing_info, environment, application_name, metrics,
    capture_message_content, disable_metrics, version, is_stream):
    """
    Process chat request and generate Telemetry
    """

    if len(scope._timestamps) > 1:
        scope._tbt = calculate_tbt(scope._timestamps)

    prompt = format_content(scope._kwargs.get("messages", []))
    request_model = scope._kwargs.get("model", "mixtral-8x7b-32768")

    cost = get_chat_model_cost(request_model, pricing_info, scope._input_tokens, scope._output_tokens)

    # Common Span Attributes
    common_span_attributes(scope,
        SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT, SemanticConvention.GEN_AI_SYSTEM_GROQ,
        scope._server_address, scope._server_port, request_model, scope._response_model,
        environment, application_name, is_stream, scope._tbt, scope._ttft, version)

    # Span Attributes for Request parameters
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_SEED, scope._kwargs.get("seed", ""))
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_FREQUENCY_PENALTY, scope._kwargs.get("frequency_penalty", 0.0))
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_MAX_TOKENS, scope._kwargs.get("max_completion_tokens", -1))
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_PRESENCE_PENALTY, scope._kwargs.get("presence_penalty", 0.0))
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_STOP_SEQUENCES, scope._kwargs.get("stop", []))
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_TEMPERATURE, scope._kwargs.get("temperature", 1.0))
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_TOP_P, scope._kwargs.get("top_p", 1.0))
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_USER, scope._kwargs.get("user", ""))
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_SERVICE_TIER, scope._kwargs.get("service_tier", "on_demand"))

    # Span Attributes for Response parameters
    scope._span.set_attribute(SemanticConvention.GEN_AI_RESPONSE_ID, scope._response_id)
    scope._span.set_attribute(SemanticConvention.GEN_AI_RESPONSE_FINISH_REASON, [scope._finish_reason])
    scope._span.set_attribute(SemanticConvention.GEN_AI_RESPONSE_SYSTEM_FINGERPRINT, scope._system_fingerprint)
    scope._span.set_attribute(SemanticConvention.GEN_AI_OUTPUT_TYPE, "text" if isinstance(scope._llmresponse, str) else "json")

    # Span Attributes for Cost and Tokens
    scope._span.set_attribute(SemanticConvention.GEN_AI_USAGE_INPUT_TOKENS, scope._input_tokens)
    scope._span.set_attribute(SemanticConvention.GEN_AI_USAGE_OUTPUT_TOKENS, scope._output_tokens)
    scope._span.set_attribute(SemanticConvention.GEN_AI_CLIENT_TOKEN_USAGE, scope._input_tokens + scope._output_tokens)
    scope._span.set_attribute(SemanticConvention.GEN_AI_USAGE_COST, cost)

    # Span Attributes for Tools
    if scope._tools:
        scope._span.set_attribute(SemanticConvention.GEN_AI_TOOL_NAME, scope._tools.get("function", {}).get("name", ""))
        scope._span.set_attribute(SemanticConvention.GEN_AI_TOOL_CALL_ID, str(scope._tools.get("id", "")))
        scope._span.set_attribute(SemanticConvention.GEN_AI_TOOL_ARGS, str(scope._tools.get("function", {}).get("arguments", "")))

    # Span Attributes for Content
    if capture_message_content:
        scope._span.set_attribute(SemanticConvention.GEN_AI_CONTENT_PROMPT, prompt)
        scope._span.set_attribute(SemanticConvention.GEN_AI_CONTENT_COMPLETION, scope._llmresponse)

        # To be removed one the change to span_attributes (from span events) is complete
        scope._span.add_event(
            name=SemanticConvention.GEN_AI_CONTENT_PROMPT_EVENT,
            attributes={
                SemanticConvention.GEN_AI_CONTENT_PROMPT: prompt,
            },
        )
        scope._span.add_event(
            name=SemanticConvention.GEN_AI_CONTENT_COMPLETION_EVENT,
            attributes={
                SemanticConvention.GEN_AI_CONTENT_COMPLETION: scope._llmresponse,
            },
        )

    scope._span.set_status(Status(StatusCode.OK))

    # Metrics
    if not disable_metrics:
        record_completion_metrics(metrics, SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT, SemanticConvention.GEN_AI_SYSTEM_GROQ,
            scope._server_address, scope._server_port, request_model, scope._response_model, environment,
            application_name, scope._start_time, scope._end_time, scope._input_tokens, scope._output_tokens,
            cost, scope._tbt, scope._ttft)

def process_streaming_chat_response(scope, pricing_info, environment, application_name, metrics,
    capture_message_content=False, disable_metrics=False, version=""):
    """
    Process chat request and generate Telemetry
    """

    common_chat_logic(scope, pricing_info, environment, application_name, metrics,
        capture_message_content, disable_metrics, version, is_stream=True)

def process_chat_response(response, request_model, pricing_info, server_port, server_address,
    environment, application_name, metrics, start_time, span, capture_message_content=False,
    disable_metrics=False, version="1.0.0", **kwargs):
    """
    Process chat request and generate Telemetry
    """

    # Create scope object
    scope = type("GenericScope", (), {})()
    response_dict = response_as_dict(response)

    scope._start_time = start_time
    scope._end_time = time.time()
    scope._span = span
    scope._llmresponse = " ".join(
        (choice.get("message", {}).get("content") or "")
        for choice in response_dict.get("choices", [])
    )
    scope._response_id = response_dict.get("id")
    scope._response_model = response_dict.get("model")
    scope._input_tokens = response_dict.get("usage", {}).get("prompt_tokens", 0)
    scope._output_tokens = response_dict.get("usage", {}).get("completion_tokens", 0)
    scope._timestamps = []
    scope._ttft, scope._tbt = scope._end_time - scope._start_time, 0
    scope._server_address, scope._server_port = server_address, server_port
    scope._kwargs = kwargs
    scope._system_fingerprint = response_dict.get("system_fingerprint")
    scope._finish_reason = str(response_dict.get("choices", [])[0].get("finish_reason", ""))

    # Handle tool calls
    if scope._kwargs.get("tools"):
        scope._tools = response_dict.get("choices", [{}])[0].get("message", {}).get("tool_calls")
    else:
        scope._tools = None

    common_chat_logic(scope, pricing_info, environment, application_name, metrics,
        capture_message_content, disable_metrics, version, is_stream=False)

    return response



# pylint: disable=useless-return, bad-staticmethod-argument, disable=duplicate-code
"""Initializer of Auto Instrumentation of Qdrant Functions"""
from typing import Collection
import importlib.metadata
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from wrapt import wrap_function_wrapper

from openlit.instrumentation.qdrant.qdrant import general_wrap
from openlit.instrumentation.qdrant.async_qdrant import async_general_wrap

_instruments = ("qdrant-client >= 1.9.0",)

WRAPPED_METHODS = [
    {
        "package": "qdrant_client",
        "object": "QdrantClient.create_collection",
        "endpoint": "qdrant.create_collection",
        "wrapper": general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "QdrantClient.delete_collection",
        "endpoint": "qdrant.delete_collection",
        "wrapper": general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "QdrantClient.update_collection",
        "endpoint": "qdrant.update_collection",
        "wrapper": general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "QdrantClient.upload_collection",
        "endpoint": "qdrant.upload_collection",
        "wrapper": general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "QdrantClient.upsert",
        "endpoint": "qdrant.upsert",
        "wrapper": general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "QdrantClient.set_payload",
        "endpoint": "qdrant.set_payload",
        "wrapper": general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "QdrantClient.overwrite_payload",
        "endpoint": "qdrant.overwrite_payload",
        "wrapper": general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "QdrantClient.clear_payload",
        "endpoint": "qdrant.clear_payload",
        "wrapper": general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "QdrantClient.delete_payload",
        "endpoint": "qdrant.delete_payload",
        "wrapper": general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "QdrantClient.upload_points",
        "endpoint": "qdrant.upload_points",
        "wrapper": general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "QdrantClient.update_vectors",
        "endpoint": "qdrant.update_vectors",
        "wrapper": general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "QdrantClient.delete_vectors",
        "endpoint": "qdrant.delete_vectors",
        "wrapper": general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "QdrantClient.delete",
        "endpoint": "qdrant.delete",
        "wrapper": general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "QdrantClient.retrieve",
        "endpoint": "qdrant.retrieve",
        "wrapper": general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "QdrantClient.scroll",
        "endpoint": "qdrant.scroll",
        "wrapper": general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "QdrantClient.search",
        "endpoint": "qdrant.search",
        "wrapper": general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "QdrantClient.search_groups",
        "endpoint": "qdrant.search_groups",
        "wrapper": general_wrap,
    },
    {

        "package": "qdrant_client",
        "object": "QdrantClient.recommend",
        "endpoint": "qdrant.recommend",
        "wrapper": general_wrap,
    },
    {

        "package": "qdrant_client",
        "object": "QdrantClient.create_payload_index",
        "endpoint": "qdrant.create_payload_index",
        "wrapper": general_wrap,
    },
    {

        "package": "qdrant_client",
        "object": "QdrantClient.query_points",
        "endpoint": "qdrant.query_points",
        "wrapper": general_wrap,
    },

    # Async Client
    {
        "package": "qdrant_client",
        "object": "AsyncQdrantClient.create_collection",
        "endpoint": "qdrant.create_collection",
        "wrapper": async_general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "AsyncQdrantClient.delete_collection",
        "endpoint": "qdrant.delete_collection",
        "wrapper": async_general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "AsyncQdrantClient.update_collection",
        "endpoint": "qdrant.update_collection",
        "wrapper": async_general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "AsyncQdrantClient.upload_collection",
        "endpoint": "qdrant.upload_collection",
        "wrapper": async_general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "AsyncQdrantClient.upsert",
        "endpoint": "qdrant.upsert",
        "wrapper": async_general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "AsyncQdrantClient.set_payload",
        "endpoint": "qdrant.set_payload",
        "wrapper": async_general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "AsyncQdrantClient.overwrite_payload",
        "endpoint": "qdrant.overwrite_payload",
        "wrapper": async_general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "AsyncQdrantClient.clear_payload",
        "endpoint": "qdrant.clear_payload",
        "wrapper": async_general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "AsyncQdrantClient.delete_payload",
        "endpoint": "qdrant.delete_payload",
        "wrapper": async_general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "AsyncQdrantClient.upload_points",
        "endpoint": "qdrant.upload_points",
        "wrapper": async_general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "AsyncQdrantClient.update_vectors",
        "endpoint": "qdrant.update_vectors",
        "wrapper": async_general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "AsyncQdrantClient.delete_vectors",
        "endpoint": "qdrant.delete_vectors",
        "wrapper": async_general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "AsyncQdrantClient.delete",
        "endpoint": "qdrant.delete",
        "wrapper": async_general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "AsyncQdrantClient.retrieve",
        "endpoint": "qdrant.retrieve",
        "wrapper": async_general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "AsyncQdrantClient.scroll",
        "endpoint": "qdrant.scroll",
        "wrapper": async_general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "AsyncQdrantClient.search",
        "endpoint": "qdrant.search",
        "wrapper": async_general_wrap,
    },
    {
        "package": "qdrant_client",
        "object": "AsyncQdrantClient.search_groups",
        "endpoint": "qdrant.search_groups",
        "wrapper": async_general_wrap,
    },
    {

        "package": "qdrant_client",
        "object": "AsyncQdrantClient.recommend",
        "endpoint": "qdrant.recommend",
        "wrapper": async_general_wrap,
    },
    {

        "package": "qdrant_client",
        "object": "AsyncQdrantClient.create_payload_index",
        "endpoint": "qdrant.create_payload_index",
        "wrapper": async_general_wrap,
    },
    {

        "package": "qdrant_client",
        "object": "AsyncQdrantClient.query_points",
        "endpoint": "qdrant.query_points",
        "wrapper": async_general_wrap,
    },
]

class QdrantInstrumentor(BaseInstrumentor):
    """An instrumentor for Qdrant's client library."""

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        application_name = kwargs.get("application_name")
        environment = kwargs.get("environment")
        tracer = kwargs.get("tracer")
        metrics = kwargs.get("metrics_dict")
        pricing_info = kwargs.get("pricing_info")
        capture_message_content = kwargs.get("capture_message_content")
        disable_metrics = kwargs.get("disable_metrics")
        version = importlib.metadata.version("qdrant-client")

        for wrapped_method in WRAPPED_METHODS:
            wrap_package = wrapped_method.get("package")
            wrap_object = wrapped_method.get("object")
            gen_ai_endpoint = wrapped_method.get("endpoint")
            wrapper = wrapped_method.get("wrapper")
            wrap_function_wrapper(
                wrap_package,
                wrap_object,
                wrapper(gen_ai_endpoint, version, environment, application_name,
                 tracer, pricing_info, capture_message_content, metrics, disable_metrics),
            )


    @staticmethod
    def _uninstrument(self, **kwargs):
        pass



"""Initializer of Auto Instrumentation of AstraDB Functions"""
from typing import Collection
import importlib.metadata
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from wrapt import wrap_function_wrapper

from openlit.instrumentation.astra.astra import general_wrap
# from openlit.instrumentation.astra.async_astra import asyc_general_wrap

_instruments = ('astrapy >= 1.5.2',)

class AstraInstrumentor(BaseInstrumentor):
    """An instrumentor for AstraDB's client library."""

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        application_name = kwargs.get('application_name')
        environment = kwargs.get('environment')
        tracer = kwargs.get('tracer')
        metrics = kwargs.get('metrics_dict')
        pricing_info = kwargs.get('pricing_info')
        capture_message_content = kwargs.get('capture_message_content')
        disable_metrics = kwargs.get('disable_metrics')
        version = importlib.metadata.version('astrapy')

        # Sync
        wrap_function_wrapper(
            'astrapy.database',
            'Database.create_collection',
            general_wrap('astra.create_collection', version, environment, application_name,
                 tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )
        wrap_function_wrapper(
            'astrapy.database',
            'Database.drop_collection',
            general_wrap('astra.drop_collection', version, environment, application_name,
                 tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )
        wrap_function_wrapper(
            'astrapy.collection',
            'Collection.insert_one',
            general_wrap('astra.insert', version, environment, application_name,
                 tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )
        wrap_function_wrapper(
            'astrapy.collection',
            'Collection.insert_many',
            general_wrap('astra.insert', version, environment, application_name,
                 tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )
        wrap_function_wrapper(
            'astrapy.collection',
            'Collection.update_one',
            general_wrap('astra.update', version, environment, application_name,
                 tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )
        wrap_function_wrapper(
            'astrapy.collection',
            'Collection.update_many',
            general_wrap('astra.update', version, environment, application_name,
                 tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )
        wrap_function_wrapper(
            'astrapy.collection',
            'Collection.find_one_and_update',
            general_wrap('astra.find_one_and_update', version, environment, application_name,
                 tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )
        wrap_function_wrapper(
            'astrapy.collection',
            'Collection.find',
            general_wrap('astra.find', version, environment, application_name,
                 tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )
        wrap_function_wrapper(
            'astrapy.collection',
            'Collection.replace_one',
            general_wrap('astra.replace_one', version, environment, application_name,
                 tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )
        wrap_function_wrapper(
            'astrapy.collection',
            'Collection.find_one_and_delete',
            general_wrap('astra.find_one_and_delete', version, environment, application_name,
                 tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )
        wrap_function_wrapper(
            'astrapy.collection',
            'Collection.delete_one',
            general_wrap('astra.delete', version, environment, application_name,
                 tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )
        wrap_function_wrapper(
            'astrapy.collection',
            'Collection.delete_many',
            general_wrap('astra.delete', version, environment, application_name,
                 tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )

        # ASync
        wrap_function_wrapper(
            'astrapy.database',
            'AsyncDatabase.create_collection',
            general_wrap('astra.create_collection', version, environment, application_name,
                 tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )
        wrap_function_wrapper(
            'astrapy.database',
            'AsyncDatabase.drop_collection',
            general_wrap('astra.drop_collection', version, environment, application_name,
                 tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )
        wrap_function_wrapper(
            'astrapy.collection',
            'AsyncCollection.insert_one',
            general_wrap('astra.insert_one', version, environment, application_name,
                 tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )
        wrap_function_wrapper(
            'astrapy.collection',
            'AsyncCollection.insert_many',
            general_wrap('astra.insert_many', version, environment, application_name,
                 tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )
        wrap_function_wrapper(
            'astrapy.collection',
            'AsyncCollection.update_one',
            general_wrap('astra.update_one', version, environment, application_name,
                 tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )
        wrap_function_wrapper(
            'astrapy.collection',
            'AsyncCollection.update_many',
            general_wrap('astra.update_many', version, environment, application_name,
                 tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )
        wrap_function_wrapper(
            'astrapy.collection',
            'AsyncCollection.find_one_and_update',
            general_wrap('astra.find_one_and_update', version, environment, application_name,
                 tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )
        wrap_function_wrapper(
            'astrapy.collection',
            'AsyncCollection.find',
            general_wrap('astra.find', version, environment, application_name,
                 tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )
        wrap_function_wrapper(
            'astrapy.collection',
            'AsyncCollection.replace_one',
            general_wrap('astra.replace_one', version, environment, application_name,
                 tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )
        wrap_function_wrapper(
            'astrapy.collection',
            'AsyncCollection.find_one_and_delete',
            general_wrap('astra.find_one_and_delete', version, environment, application_name,
                 tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )
        wrap_function_wrapper(
            'astrapy.collection',
            'AsyncCollection.delete_one',
            general_wrap('astra.delete_one', version, environment, application_name,
                 tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )
        wrap_function_wrapper(
            'astrapy.collection',
            'AsyncCollection.delete_many',
            general_wrap('astra.delete_many', version, environment, application_name,
                 tracer, pricing_info, capture_message_content, metrics, disable_metrics),
        )

    def _uninstrument(self, **kwargs):
        # Proper uninstrumentation logic to revert patched methods
        pass



"""
Module for monitoring AG2 API calls.
"""

import logging
import time
from opentelemetry.trace import SpanKind, Status, StatusCode
from opentelemetry.sdk.resources import SERVICE_NAME, TELEMETRY_SDK_NAME, DEPLOYMENT_ENVIRONMENT
from openlit.__helpers import (
    handle_exception,
    get_chat_model_cost,
    otel_event,
)
from openlit.semcov import SemanticConvention

# Initialize logger for logging potential issues and operations
logger = logging.getLogger(__name__)

AGENT_NAME = ''
REQUEST_MODEL = ''
SYSTEM_MESSAGE = ''
MODEL_AND_NAME_SET = False

def set_span_attributes(span, version, operation_name, environment,
        application_name, server_address, server_port, request_model):
    """
    Set common attributes for the span.
    """

    # Set Span attributes (OTel Semconv)
    span.set_attribute(TELEMETRY_SDK_NAME, 'openlit')
    span.set_attribute(SemanticConvention.GEN_AI_OPERATION, operation_name)
    span.set_attribute(SemanticConvention.GEN_AI_SYSTEM, SemanticConvention.GEN_AI_SYSTEM_AG2)
    span.set_attribute(SemanticConvention.GEN_AI_AGENT_NAME, AGENT_NAME)
    span.set_attribute(SemanticConvention.SERVER_ADDRESS, server_address)
    span.set_attribute(SemanticConvention.SERVER_PORT, server_port)
    span.set_attribute(SemanticConvention.GEN_AI_REQUEST_MODEL, request_model)

    # Set Span attributes (Extras)
    span.set_attribute(DEPLOYMENT_ENVIRONMENT, environment)
    span.set_attribute(SERVICE_NAME, application_name)
    span.set_attribute(SemanticConvention.GEN_AI_SDK_VERSION, version)

def calculate_tokens_and_cost(response, request_model, pricing_info):
    """
    Calculate the input, output tokens, and their respective costs.
    """
    input_tokens = 0
    output_tokens = 0

    for usage_data in response.cost.values():
        if isinstance(usage_data, dict):
            for model_data in usage_data.values():
                if isinstance(model_data, dict):
                    input_tokens += model_data.get('prompt_tokens', 0)
                    output_tokens += model_data.get('completion_tokens', 0)

    cost = get_chat_model_cost(request_model, pricing_info, input_tokens, output_tokens)
    return input_tokens, output_tokens, cost

def emit_events(response, event_provider, capture_message_content):
    """
    Emit OpenTelemetry events for each chat history entry.
    """
    for chat in response.chat_history:
        event_type = (
            SemanticConvention.GEN_AI_CHOICE if chat['role'] == 'user'
            else SemanticConvention.GEN_AI_USER_MESSAGE
        )
        choice_event = otel_event(
            name=event_type,
            attributes={
                SemanticConvention.GEN_AI_SYSTEM: SemanticConvention.GEN_AI_SYSTEM_AG2
            },
            body={
                'index': response.chat_history.index(chat),
                'message': {
                    **({'content': chat['content']} if capture_message_content else {}),
                    'role': 'assistant' if chat['role'] == 'user' else 'user'
                }
            }
        )
        event_provider.emit(choice_event)

def conversable_agent(version, environment, application_name,
                      tracer, event_provider, pricing_info, capture_message_content, metrics, disable_metrics):
    """
    Generates a telemetry wrapper for GenAI function call
    """
    def wrapper(wrapped, instance, args, kwargs):
        server_address, server_port = '127.0.0.1', 80
        global AGENT_NAME, MODEL_AND_NAME_SET, REQUEST_MODEL, SYSTEM_MESSAGE

        if not MODEL_AND_NAME_SET:
            AGENT_NAME = kwargs.get('name', 'NOT_FOUND')
            REQUEST_MODEL = kwargs.get('llm_config', {}).get('model', 'gpt-4o')
            SYSTEM_MESSAGE = kwargs.get('system_message', '')
            MODEL_AND_NAME_SET = True

        span_name = f'{SemanticConvention.GEN_AI_OPERATION_TYPE_CREATE_AGENT} {AGENT_NAME}'

        with tracer.start_as_current_span(span_name, kind=SpanKind.CLIENT) as span:
            try:
                start_time = time.time()
                response = wrapped(*args, **kwargs)
                end_time = time.time()

                set_span_attributes(span, version, SemanticConvention.GEN_AI_OPERATION_TYPE_CREATE_AGENT,
                    environment, application_name, server_address, server_port, REQUEST_MODEL)
                span.set_attribute(SemanticConvention.GEN_AI_AGENT_DESCRIPTION, SYSTEM_MESSAGE)
                span.set_attribute(SemanticConvention.GEN_AI_RESPONSE_MODEL, REQUEST_MODEL)
                span.set_attribute(SemanticConvention.GEN_AI_SERVER_TTFT, end_time - start_time)

                span.set_status(Status(StatusCode.OK))

                return response

            except Exception as e:
                handle_exception(span, e)
                logger.error('Error in trace creation: %s', e)
                return response

    return wrapper

def agent_run(version, environment, application_name,
              tracer, event_provider, pricing_info, capture_message_content, metrics, disable_metrics):
    """
    Generates a telemetry wrapper for GenAI function call
    """
    def wrapper(wrapped, instance, args, kwargs):
        server_address, server_port = '127.0.0.1', 80

        span_name = f'{SemanticConvention.GEN_AI_OPERATION_TYPE_EXECUTE_AGENT_TASK} {AGENT_NAME}'

        with tracer.start_as_current_span(span_name, kind=SpanKind.CLIENT) as span:
            try:
                start_time = time.time()
                response = wrapped(*args, **kwargs)
                end_time = time.time()

                input_tokens, output_tokens, cost = calculate_tokens_and_cost(response, REQUEST_MODEL, pricing_info)
                response_model = list(response.cost.get('usage_including_cached_inference', {}).keys())[1]

                set_span_attributes(span, version, SemanticConvention.GEN_AI_OPERATION_TYPE_EXECUTE_AGENT_TASK,
                    environment, application_name, server_address, server_port, REQUEST_MODEL)
                span.set_attribute(SemanticConvention.GEN_AI_RESPONSE_MODEL, response_model)
                span.set_attribute(SemanticConvention.GEN_AI_USAGE_INPUT_TOKENS, input_tokens)
                span.set_attribute(SemanticConvention.GEN_AI_USAGE_OUTPUT_TOKENS, output_tokens)
                span.set_attribute(SemanticConvention.GEN_AI_CLIENT_TOKEN_USAGE, input_tokens + output_tokens)
                span.set_attribute(SemanticConvention.GEN_AI_USAGE_COST, cost)
                span.set_attribute(SemanticConvention.GEN_AI_SERVER_TTFT, end_time - start_time)

                emit_events(response, event_provider, capture_message_content)
                span.set_status(Status(StatusCode.OK))

                return response

            except Exception as e:
                handle_exception(span, e)
                logger.error('Error in trace creation: %s', e)
                return response

    return wrapper



"""Initializer of Auto Instrumentation of OpenAI Agents Functions"""

from typing import Collection
import importlib.metadata
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from wrapt import wrap_function_wrapper

from openlit.instrumentation.openai_agents.openai_agents import (
    create_agent
)

_instruments = ('openai-agents >= 0.0.3',)

class OpenAIAgentsInstrumentor(BaseInstrumentor):
    """
    An instrumentor for openai-agents's client library.
    """

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        application_name = kwargs.get('application_name', 'default')
        environment = kwargs.get('environment', 'default')
        tracer = kwargs.get('tracer')
        event_provider = kwargs.get('event_provider')
        metrics = kwargs.get('metrics_dict')
        pricing_info = kwargs.get('pricing_info', {})
        capture_message_content = kwargs.get('capture_message_content', False)
        disable_metrics = kwargs.get('disable_metrics')
        version = importlib.metadata.version('openai-agents')

        wrap_function_wrapper(
            'agents.agent',
            'Agent.__init__',
            create_agent(version, environment, application_name,
                  tracer, event_provider, pricing_info, capture_message_content, metrics, disable_metrics),
        )

    def _uninstrument(self, **kwargs):
        # Proper uninstrumentation logic to revert patched methods
        pass



"""
Module for monitoring Pydantic AI API calls.
"""

from openlit.instrumentation.pydantic_ai.utils import (
    common_agent_run,
    common_agent_create
)

def agent_create(version, environment, application_name,
    tracer, pricing_info, capture_message_content, metrics, disable_metrics):

    """
    Generates a telemetry wrapper for GenAI function call
    """

    def wrapper(wrapped, instance, args, kwargs):
        response = wrapped(*args, **kwargs)
        return common_agent_create(wrapped, instance, args, kwargs, tracer,
                                 version, environment, application_name,
                                 capture_message_content, response=response)

    return wrapper

def agent_run(version, environment, application_name,
              tracer, pricing_info, capture_message_content, metrics, disable_metrics):
    """
    Generates a telemetry wrapper for GenAI function call
    """

    def wrapper(wrapped, instance, args, kwargs):
        response = wrapped(*args, **kwargs)
        return common_agent_run(wrapped, instance, args, kwargs, tracer,
                                    version, environment, application_name,
                                    capture_message_content, response=response)

    return wrapper

def async_agent_run(version, environment, application_name,
                    tracer, pricing_info, capture_message_content, metrics, disable_metrics):
    """
    Generates a telemetry wrapper for GenAI function call
    """

    async def wrapper(wrapped, instance, args, kwargs):
        response = await wrapped(*args, **kwargs)
        return common_agent_run(wrapped, instance, args, kwargs, tracer,
                                          version, environment, application_name,
                                          capture_message_content, response=response)

    return wrapper



"""Initializer of Auto Instrumentation of Azure AI Inference Functions"""

from typing import Collection
import importlib.metadata
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from wrapt import wrap_function_wrapper
from openlit.instrumentation.azure_ai_inference.azure_ai_inference import (
    complete
)
from openlit.instrumentation.azure_ai_inference.async_azure_ai_inference import (
    async_complete
)

_instruments = ('azure-ai-inference >= 1.0.0b4',)

class AzureAIInferenceInstrumentor(BaseInstrumentor):
    """
    An instrumentor for azure-ai-inference's client library.
    """

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        application_name = kwargs.get('application_name', 'default')
        environment = kwargs.get('environment', 'default')
        tracer = kwargs.get('tracer')
        event_provider = kwargs.get('event_provider')
        metrics = kwargs.get('metrics_dict')
        pricing_info = kwargs.get('pricing_info', {})
        capture_message_content = kwargs.get('capture_message_content', False)
        disable_metrics = kwargs.get('disable_metrics')
        version = importlib.metadata.version('azure-ai-inference')

        # sync generate
        wrap_function_wrapper(
            'azure.ai.inference',
            'ChatCompletionsClient.complete',
            complete(version, environment, application_name,
                  tracer, event_provider, pricing_info, capture_message_content, metrics, disable_metrics),
        )

        # async generate
        wrap_function_wrapper(
            'azure.ai.inference.aio',
            'ChatCompletionsClient.complete',
            async_complete(version, environment, application_name,
                  tracer, event_provider, pricing_info, capture_message_content, metrics, disable_metrics),
        )

    def _uninstrument(self, **kwargs):
        # Proper uninstrumentation logic to revert patched methods
        pass



"""
Module for monitoring Anthropic API calls.
"""

import logging
import time
from opentelemetry.trace import SpanKind
from openlit.__helpers import (
    handle_exception,
    set_server_address_and_port
)
from openlit.instrumentation.anthropic.utils import (
    process_chunk,
    process_chat_response,
    process_streaming_chat_response,
)
from openlit.semcov import SemanticConvention

# Initialize logger for logging potential issues and operations
logger = logging.getLogger(__name__)

def messages(version, environment, application_name, tracer, event_provider,
             pricing_info, capture_message_content, metrics, disable_metrics):
    """
    Generates a telemetry wrapper for GenAI function call
    """

    class TracedSyncStream:
        """
        Wrapper for streaming responses to collect telemetry.
        """

        def __init__(
                self,
                wrapped,
                span,
                span_name,
                kwargs,
                server_address,
                server_port,
                **args,
            ):
            self.__wrapped__ = wrapped
            self._span = span
            self._span_name = span_name
            self._llmresponse = ''
            self._response_id = ''
            self._response_model = ''
            self._finish_reason = ''
            self._input_tokens = ''
            self._output_tokens = ''
            self._tool_arguments = ''
            self._tool_id = ''
            self._tool_name = ''
            self._tool_calls = None
            self._response_role = ''

            self._args = args
            self._kwargs = kwargs
            self._start_time = time.time()
            self._end_time = None
            self._timestamps = []
            self._ttft = 0
            self._tbt = 0
            self._server_address = server_address
            self._server_port = server_port

        def __enter__(self):
            self.__wrapped__.__enter__()
            return self

        def __exit__(self, exc_type, exc_value, traceback):
            self.__wrapped__.__exit__(exc_type, exc_value, traceback)

        def __iter__(self):
            return self

        def __getattr__(self, name):
            """Delegate attribute access to the wrapped object."""
            return getattr(self.__wrapped__, name)

        def __next__(self):
            try:
                chunk = self.__wrapped__.__next__()
                process_chunk(self, chunk)
                return chunk
            except StopIteration:
                try:
                    with tracer.start_as_current_span(self._span_name, kind= SpanKind.CLIENT) as self._span:
                        process_streaming_chat_response(
                            self,
                            pricing_info=pricing_info,
                            environment=environment,
                            application_name=application_name,
                            metrics=metrics,
                            event_provider=event_provider,
                            capture_message_content=capture_message_content,
                            disable_metrics=disable_metrics,
                            version=version
                        )

                except Exception as e:
                    handle_exception(self._span, e)
                    logger.error("Error in trace creation: %s", e)
                raise

    def wrapper(wrapped, instance, args, kwargs):
        """
        Wraps the GenAI function call.
        """

        streaming = kwargs.get('stream', False)
        server_address, server_port = set_server_address_and_port(instance, 'api.anthropic.com', 443)
        request_model = kwargs.get('model', 'claude-3-5-sonnet-latest')

        span_name = f'{SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT} {request_model}'

        # pylint: disable=no-else-return
        if streaming:
            awaited_wrapped = wrapped(*args, **kwargs)
            span = tracer.start_span(span_name, kind=SpanKind.CLIENT)

            return TracedSyncStream(awaited_wrapped, span, span_name, kwargs, server_address, server_port)

        else:
            with tracer.start_as_current_span(span_name, kind=SpanKind.CLIENT) as span:
                start_time = time.time()
                response = wrapped(*args, **kwargs)
                response = process_chat_response(
                    response=response,
                    request_model=request_model,
                    pricing_info=pricing_info,
                    server_port=server_port,
                    server_address=server_address,
                    environment=environment,
                    application_name=application_name,
                    metrics=metrics,
                    event_provider=event_provider,
                    start_time=start_time,
                    span=span,
                    capture_message_content=capture_message_content,
                    disable_metrics=disable_metrics,
                    version=version,
                    **kwargs
                )

            return response

    return wrapper



"""
Module for monitoring AstraDB.
"""

import time
from opentelemetry.trace import SpanKind
from openlit.instrumentation.astra.utils import (
    DB_OPERATION_MAP,
    process_db_operations
)
from openlit.semcov import SemanticConvention

def general_wrap(gen_ai_endpoint, version, environment, application_name,
                 tracer, pricing_info, capture_message_content, metrics, disable_metrics):
    """
    Generates a telemetry wrapper for VectorDB function call
    """

    def wrapper(wrapped, instance, args, kwargs):
        """
        Wraps the VectorDB function call.
        """

        db_operation = DB_OPERATION_MAP.get(gen_ai_endpoint, "UNKNOWN")
        if db_operation == SemanticConvention.DB_OPERATION_REPLACE and kwargs.get('upsert'):
            db_operation = SemanticConvention.DB_OPERATION_UPSERT

        span_name = f"{db_operation} {instance.name}"

        with tracer.start_as_current_span(span_name, kind=SpanKind.CLIENT) as span:
            start_time = time.time()
            response = wrapped(*args, **kwargs)
            server_address = getattr(getattr(instance, 'database', instance), 'api_endpoint', '')
            server_port = 443
            collection_name = instance.name
            response = process_db_operations(
                response, span, start_time, gen_ai_endpoint,
                version, environment, application_name, capture_message_content,
                metrics, disable_metrics, server_address, server_port,
                collection_name, db_operation, kwargs, args
            )

            return response

    return wrapper



"""
Together AI OpenTelemetry instrumentation utility functions
"""
import time

from opentelemetry.sdk.resources import SERVICE_NAME, TELEMETRY_SDK_NAME, DEPLOYMENT_ENVIRONMENT
from opentelemetry.trace import Status, StatusCode

from openlit.__helpers import (
    calculate_ttft,
    response_as_dict,
    calculate_tbt,
    get_chat_model_cost,
    get_image_model_cost,
    create_metrics_attributes,
)
from openlit.semcov import SemanticConvention

def format_content(messages):
    """
    Process a list of messages to extract content.
    """

    formatted_messages = []
    for message in messages:
        role = message["role"]
        content = message["content"]

        if isinstance(content, list):
            content_str = ", ".join(
                f'{item["type"]}: {item["text"] if "text" in item else item["image_url"]}'
                if "type" in item else f'text: {item["text"]}'
                for item in content
            )
            formatted_messages.append(f"{role}: {content_str}")
        else:
            formatted_messages.append(f"{role}: {content}")

    return "\n".join(formatted_messages)

def process_chunk(scope, chunk):
    """
    Process a chunk of response data and update state.
    """

    end_time = time.time()
    # Record the timestamp for the current chunk
    scope._timestamps.append(end_time)

    if len(scope._timestamps) == 1:
        # Calculate time to first chunk
        scope._ttft = calculate_ttft(scope._timestamps, scope._start_time)

    chunked = response_as_dict(chunk)
    # Collect message IDs and aggregated response from events
    if (len(chunked.get("choices")) > 0 and ("delta" in chunked.get("choices")[0] and
        "content" in chunked.get("choices")[0].get("delta"))):

        content = chunked.get("choices")[0].get("delta").get("content")
        if content:
            scope._llmresponse += content

    if chunked.get("usage"):
        scope._response_id = chunked.get("id")
        scope._response_model = chunked.get("model")
        scope._input_tokens = chunked.get("usage").get("prompt_tokens")
        scope._output_tokens = chunked.get("usage").get("completion_tokens")
        scope._finish_reason = str(chunked.get("finish_reason"))
        scope._end_time = time.time()

def common_span_attributes(scope, gen_ai_operation, gen_ai_system, server_address, server_port,
    request_model, response_model, environment, application_name, is_stream, tbt, ttft, version):
    """
    Set common span attributes for both chat and RAG operations.
    """

    scope._span.set_attribute(TELEMETRY_SDK_NAME, "openlit")
    scope._span.set_attribute(SemanticConvention.GEN_AI_OPERATION, gen_ai_operation)
    scope._span.set_attribute(SemanticConvention.GEN_AI_SYSTEM, gen_ai_system)
    scope._span.set_attribute(SemanticConvention.SERVER_ADDRESS, server_address)
    scope._span.set_attribute(SemanticConvention.SERVER_PORT, server_port)
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_MODEL, request_model)
    scope._span.set_attribute(SemanticConvention.GEN_AI_RESPONSE_MODEL, response_model)
    scope._span.set_attribute(DEPLOYMENT_ENVIRONMENT, environment)
    scope._span.set_attribute(SERVICE_NAME, application_name)
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_IS_STREAM, is_stream)
    scope._span.set_attribute(SemanticConvention.GEN_AI_SERVER_TBT, tbt)
    scope._span.set_attribute(SemanticConvention.GEN_AI_SERVER_TTFT, ttft)
    scope._span.set_attribute(SemanticConvention.GEN_AI_SDK_VERSION, version)

def record_common_metrics(metrics, gen_ai_operation, gen_ai_system, server_address, server_port,
    request_model, response_model, environment, application_name, start_time, end_time,
    input_tokens, output_tokens, cost, tbt=None, ttft=None):
    """
    Record common metrics for the operation.
    """

    attributes = create_metrics_attributes(
        operation=gen_ai_operation,
        system=gen_ai_system,
        server_address=server_address,
        server_port=server_port,
        request_model=request_model,
        response_model=response_model,
        service_name=application_name,
        deployment_environment=environment,
    )
    metrics["genai_client_operation_duration"].record(end_time - start_time, attributes)
    metrics["genai_requests"].add(1, attributes)
    metrics["genai_prompt_tokens"].add(input_tokens, attributes)
    metrics["genai_completion_tokens"].add(output_tokens, attributes)
    metrics["genai_client_usage_tokens"].record(input_tokens + output_tokens, attributes)
    metrics["genai_cost"].record(cost, attributes)
    if tbt is not None:
        metrics["genai_server_tbt"].record(tbt, attributes)
    if ttft is not None:
        metrics["genai_server_ttft"].record(ttft, attributes)

def common_chat_logic(scope, pricing_info, environment, application_name, metrics,
    capture_message_content, disable_metrics, version, is_stream):
    """
    Process chat request and generate Telemetry
    """

    if len(scope._timestamps) > 1:
        scope._tbt = calculate_tbt(scope._timestamps)

    prompt = format_content(scope._kwargs.get("messages", ""))
    request_model = scope._kwargs.get("model", "jamba-1.5-mini")

    cost = get_chat_model_cost(request_model, pricing_info, scope._input_tokens, scope._output_tokens)

    # Common Span Attributes
    common_span_attributes(scope,
        SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT,SemanticConvention.GEN_AI_SYSTEM_TOGETHER,
        scope._server_address, scope._server_port, request_model, scope._response_model,
        environment, application_name, is_stream, scope._tbt, scope._ttft, version)

    # Span Attributes for Response parameters
    scope._span.set_attribute(SemanticConvention.GEN_AI_RESPONSE_FINISH_REASON, [scope._finish_reason])
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_SEED, scope._kwargs.get("seed", ""))
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_FREQUENCY_PENALTY, scope._kwargs.get("frequency_penalty", 0.0))
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_MAX_TOKENS, scope._kwargs.get("max_tokens", -1))
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_PRESENCE_PENALTY, scope._kwargs.get("presence_penalty", 0.0))
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_STOP_SEQUENCES, scope._kwargs.get("stop", []))
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_TEMPERATURE, scope._kwargs.get("temperature", 1.0))
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_TOP_P, scope._kwargs.get("top_p", 1.0))
    scope._span.set_attribute(SemanticConvention.GEN_AI_RESPONSE_ID, scope._response_id)
    scope._span.set_attribute(SemanticConvention.GEN_AI_OUTPUT_TYPE, "text" if isinstance(scope._llmresponse, str) else "json")

    # Span Attributes for Cost and Tokens
    scope._span.set_attribute(SemanticConvention.GEN_AI_USAGE_INPUT_TOKENS, scope._input_tokens)
    scope._span.set_attribute(SemanticConvention.GEN_AI_USAGE_OUTPUT_TOKENS, scope._output_tokens)
    scope._span.set_attribute(SemanticConvention.GEN_AI_CLIENT_TOKEN_USAGE, scope._input_tokens + scope._output_tokens)
    scope._span.set_attribute(SemanticConvention.GEN_AI_USAGE_COST, cost)

    # Span Attributes for Tools
    if scope._tools:
        scope._span.set_attribute(SemanticConvention.GEN_AI_TOOL_NAME, scope._tools.get("function","")).get("name","")
        scope._span.set_attribute(SemanticConvention.GEN_AI_TOOL_CALL_ID, str(scope._tools.get("id","")))
        scope._span.set_attribute(SemanticConvention.GEN_AI_TOOL_ARGS, str(scope._tools.get("function","").get("arguments","")))

    # Span Attributes for Content
    if capture_message_content:
        scope._span.set_attribute(SemanticConvention.GEN_AI_CONTENT_PROMPT, prompt)
        scope._span.set_attribute(SemanticConvention.GEN_AI_CONTENT_COMPLETION, scope._llmresponse)

        # To be removed one the change to span_attributes (from span events) is complete
        scope._span.add_event(
            name=SemanticConvention.GEN_AI_CONTENT_PROMPT_EVENT,
            attributes={
                SemanticConvention.GEN_AI_CONTENT_PROMPT: prompt,
            },
        )
        scope._span.add_event(
            name=SemanticConvention.GEN_AI_CONTENT_COMPLETION_EVENT,
            attributes={
                SemanticConvention.GEN_AI_CONTENT_COMPLETION: scope._llmresponse,
            },
        )

    scope._span.set_status(Status(StatusCode.OK))

    # Metrics
    if not disable_metrics:
        record_common_metrics(metrics, SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT, SemanticConvention.GEN_AI_SYSTEM_TOGETHER,
            scope._server_address, scope._server_port, request_model, scope._response_model, environment,
            application_name, scope._start_time, scope._end_time, scope._input_tokens, scope._output_tokens,
            cost, scope._tbt, scope._ttft)

def process_streaming_chat_response(scope, pricing_info, environment, application_name, metrics,
    capture_message_content=False, disable_metrics=False, version=""):
    """
    Process chat request and generate Telemetry
    """
    common_chat_logic(scope, pricing_info, environment, application_name, metrics,
        capture_message_content, disable_metrics, version, is_stream=True)

def process_chat_response(response, request_model, pricing_info, server_port, server_address,
    environment, application_name, metrics, start_time, span, capture_message_content=False,
    disable_metrics=False, version="1.0.0", **kwargs):
    """
    Process chat request and generate Telemetry
    """

    scope = type("GenericScope", (), {})()
    response_dict = response_as_dict(response)

    scope._start_time = start_time
    scope._end_time = time.time()
    scope._span = span
    scope._llmresponse = " ".join(
        (choice.get("message", {}).get("content") or "")
        for choice in response_dict.get("choices", [])
    )
    scope._response_id = response_dict.get("id")
    scope._response_model = response_dict.get("model")
    scope._input_tokens = response_dict.get("usage").get("prompt_tokens")
    scope._output_tokens = response_dict.get("usage").get("completion_tokens")
    scope._timestamps = []
    scope._ttft, scope._tbt = scope._end_time - scope._start_time, 0
    scope._server_address, scope._server_port = server_address, server_port
    scope._kwargs = kwargs
    scope._finish_reason = str(response_dict.get("choices")[0].get("finish_reason"))

    if scope._kwargs.get("tools"):
        scope._tools = response_dict.get("choices")[0].get("message").get("tool_calls")
    else:
        scope._tools = None

    common_chat_logic(scope, pricing_info, environment, application_name, metrics,
        capture_message_content, disable_metrics, version, is_stream=False)

    return response

def common_image_logic(scope, pricing_info, environment, application_name, metrics,
    capture_message_content, disable_metrics, version):
    """
    Process image generation request and generate Telemetry
    """

    # Find Image format
    if "response_format" in scope._kwargs and scope._kwargs["response_format"] == "b64_json":
        image_format = "b64_json"
    else:
        image_format = "url"

    image_size = str(scope._kwargs.get("width", "1024")) + "x" + str(scope._kwargs.get("height", "1024"))
    request_model = scope._kwargs.get("model", "dall-e-2")

    # Calculate cost of the operation
    cost = get_image_model_cost(request_model, pricing_info, image_size,
        scope._kwargs.get("quality", "standard"))

    # Common Span Attributes
    common_span_attributes(scope,
        SemanticConvention.GEN_AI_OPERATION_TYPE_IMAGE, SemanticConvention.GEN_AI_SYSTEM_TOGETHER,
        scope._server_address, scope._server_port, request_model, scope._response_model,
        environment, application_name, False, scope._tbt, scope._ttft, version)

    # Image-specific span attributes
    scope._span.set_attribute(SemanticConvention.GEN_AI_RESPONSE_ID, scope._response_id)
    scope._span.set_attribute(SemanticConvention.GEN_AI_OUTPUT_TYPE, "image")
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_IMAGE_SIZE, image_size)
    scope._span.set_attribute(SemanticConvention.GEN_AI_USAGE_COST, len(scope._response_data) * cost)

    # Content attributes
    if capture_message_content:
        scope._span.add_event(
            name=SemanticConvention.GEN_AI_CONTENT_PROMPT_EVENT,
            attributes={
                SemanticConvention.GEN_AI_CONTENT_PROMPT: scope._kwargs.get("prompt", ""),
            },
        )

        for images_count, item in enumerate(scope._response_data):
            attribute_name = f"{SemanticConvention.GEN_AI_RESPONSE_IMAGE}.{images_count}"
            scope._span.add_event(
                name=attribute_name,
                attributes={
                    SemanticConvention.GEN_AI_CONTENT_COMPLETION: getattr(item, image_format),
                },
            )

    scope._span.set_status(Status(StatusCode.OK))

    # Metrics
    if not disable_metrics:
        record_common_metrics(
            metrics, SemanticConvention.GEN_AI_OPERATION_TYPE_IMAGE,
            SemanticConvention.GEN_AI_SYSTEM_TOGETHER, scope._server_address, scope._server_port,
            request_model, scope._response_model, environment, application_name,
            scope._start_time, scope._end_time, 0, 0, len(scope._response_data) * cost
        )

def process_image_response(response, request_model, pricing_info, server_address, server_port,
    environment, application_name, metrics, start_time, end_time, span, capture_message_content,
    disable_metrics, version, **kwargs):
    """
    Process image generation request and generate Telemetry
    """

    scope = type("GenericScope", (), {})()

    scope._start_time = start_time
    scope._end_time = end_time
    scope._span = span
    scope._response_id = response.id
    scope._response_model = response.model
    scope._response_data = response.data
    scope._server_address = server_address
    scope._server_port = server_port
    scope._kwargs = kwargs
    scope._tbt = 0
    scope._ttft = end_time - start_time

    common_image_logic(scope, pricing_info, environment, application_name, metrics,
        capture_message_content, disable_metrics, version)

    return response



# pylint: disable=duplicate-code, broad-exception-caught, too-many-statements, unused-argument, unused-import, too-many-function-args
"""
Module for monitoring Langchain applications.
"""

import logging
import time
from opentelemetry.trace import SpanKind, Status, StatusCode
from opentelemetry.sdk.resources import SERVICE_NAME, TELEMETRY_SDK_NAME, DEPLOYMENT_ENVIRONMENT
from openlit.__helpers import (
    get_chat_model_cost,
    handle_exception,
    general_tokens,
    calculate_ttft,
    calculate_tbt,
    create_metrics_attributes,
)
from openlit.semcov import SemanticConvention

# Initialize logger for logging potential issues and operations
logger = logging.getLogger(__name__)

def get_attribute_from_instance_or_kwargs(instance, attribute_name, default=-1):
    """Return attribute from instance or kwargs"""
    # Attempt to retrieve model_kwargs from the instance
    model_kwargs = getattr(instance, 'model_kwargs', None)

    # Check for attribute in model_kwargs if it exists
    if model_kwargs and attribute_name in model_kwargs:
        return model_kwargs[attribute_name]

    # Attempt to get the attribute directly from the instance
    try:
        return getattr(instance, attribute_name)
    except AttributeError:
        # Special handling for 'model' attribute to consider 'model_id'
        if attribute_name == 'model':
            return getattr(instance, 'model_id', 'default_model_id')

        # Default if the attribute isn't found in model_kwargs or the instance
        return default

def async_general_wrap(gen_ai_endpoint, version, environment, application_name,
                 tracer, pricing_info, capture_message_content, metrics, disable_metrics):
    """
    Creates a wrapper around a function call to trace and log its execution metrics.

    This function wraps any given function to measure its execution time,
    log its operation, and trace its execution using OpenTelemetry.
    
    Parameters:
    - gen_ai_endpoint (str): A descriptor or name for the endpoint being traced.
    - version (str): The version of the Langchain application.
    - environment (str): The deployment environment (e.g., 'production', 'development').
    - application_name (str): Name of the Langchain application.
    - tracer (opentelemetry.trace.Tracer): The tracer object used for OpenTelemetry tracing.
    - pricing_info (dict): Information about the pricing for internal metrics (currently not used).
    - capture_message_content (bool): Flag indicating whether to trace the content of the response.

    Returns:
    - function: A higher-order function that takes a function 'wrapped' and returns
                a new function that wraps 'wrapped' with additional tracing and logging.
    """

    async def wrapper(wrapped, instance, args, kwargs):
        """
        An inner wrapper function that executes the wrapped function, measures execution
        time, and records trace data using OpenTelemetry.

        Parameters:
        - wrapped (Callable): The original function that this wrapper will execute.
        - instance (object): The instance to which the wrapped function belongs. This
                             is used for instance methods. For static and classmethods,
                             this may be None.
        - args (tuple): Positional arguments passed to the wrapped function.
        - kwargs (dict): Keyword arguments passed to the wrapped function.

        Returns:
        - The result of the wrapped function call.
        
        The wrapper initiates a span with the provided tracer, sets various attributes
        on the span based on the function's execution and response, and ensures
        errors are handled and logged appropriately.
        """
        with tracer.start_as_current_span(gen_ai_endpoint, kind= SpanKind.CLIENT) as span:
            response = await wrapped(*args, **kwargs)

            try:
                span.set_attribute(TELEMETRY_SDK_NAME, "openlit")
                span.set_attribute(SemanticConvention.GEN_AI_SYSTEM,
                                    SemanticConvention.GEN_AI_SYSTEM_LANGCHAIN)
                span.set_attribute(SemanticConvention.GEN_AI_ENDPOINT,
                                    gen_ai_endpoint)
                span.set_attribute(DEPLOYMENT_ENVIRONMENT,
                                    environment)
                span.set_attribute(SemanticConvention.GEN_AI_OPERATION,
                                    SemanticConvention.GEN_AI_OPERATION_TYPE_FRAMEWORK)
                span.set_attribute(SERVICE_NAME,
                                    application_name)
                span.set_attribute(SemanticConvention.GEN_AI_RETRIEVAL_SOURCE,
                                    response[0].metadata["source"])
                span.set_status(Status(StatusCode.OK))

                # Return original response
                return response

            except Exception as e:
                handle_exception(span, e)
                logger.error("Error in trace creation: %s", e)

                # Return original response
                return response

    return wrapper

def async_hub(gen_ai_endpoint, version, environment, application_name, tracer,
        pricing_info, capture_message_content, metrics, disable_metrics):
    """
    Creates a wrapper around Langchain hub operations for tracing and logging.

    Similar to `general_wrap`, this function focuses on wrapping functions involved
    in interacting with the Langchain hub, adding specific metadata relevant to
    hub operations to the span attributes.

    Parameters:
    - gen_ai_endpoint (str): A descriptor or name for the Langchain hub endpoint.
    - version (str): The version of the Langchain application.
    - environment (str): The deployment environment, such as 'production' or 'development'.
    - application_name (str): Name of the Langchain application.
    - tracer (opentelemetry.trace.Tracer): The tracer for OpenTelemetry tracing.
    - pricing_info (dict): Pricing information for the operation (not currently used).
    - capture_message_content (bool): Indicates if the content of the response should be traced.

    Returns:
    - function: A new function that wraps the original hub operation call with added
                logging, tracing, and metric calculation functionalities.
    """

    async def wrapper(wrapped, instance, args, kwargs):
        """
        An inner wrapper specifically designed for Langchain hub operations,
        providing tracing, logging, and execution metrics.

        Parameters:
        - wrapped (Callable): The original hub operation function to be executed.
        - instance (object): The instance of the class where the hub operation
                             method is defined. May be None for static or class methods.
        - args (tuple): Positional arguments to pass to the hub operation function.
        - kwargs (dict): Keyword arguments to pass to the hub operation function.

        Returns:
        - The result of executing the hub operation function.
        
        This wrapper captures additional metadata relevant to Langchain hub operations,
        creating spans with specific attributes and metrics that reflect the nature of
        each hub call.
        """

        with tracer.start_as_current_span(gen_ai_endpoint, kind= SpanKind.CLIENT) as span:
            response = await wrapped(*args, **kwargs)

            try:
                span.set_attribute(TELEMETRY_SDK_NAME, "openlit")
                span.set_attribute(SemanticConvention.GEN_AI_SYSTEM,
                                    SemanticConvention.GEN_AI_SYSTEM_LANGCHAIN)
                span.set_attribute(SemanticConvention.GEN_AI_ENDPOINT,
                                    gen_ai_endpoint)
                span.set_attribute(DEPLOYMENT_ENVIRONMENT,
                                    environment)
                span.set_attribute(SemanticConvention.GEN_AI_OPERATION,
                                    SemanticConvention.GEN_AI_OPERATION_TYPE_FRAMEWORK)
                span.set_attribute(SERVICE_NAME,
                                    application_name)
                span.set_attribute(SemanticConvention.GEN_AI_HUB_OWNER,
                                    response.metadata["lc_hub_owner"])
                span.set_attribute(SemanticConvention.GEN_AI_HUB_REPO,
                                    response.metadata["lc_hub_repo"])
                span.set_status(Status(StatusCode.OK))

                return response

            except Exception as e:
                handle_exception(span, e)
                logger.error("Error in trace creation: %s", e)

                # Return original response
                return response

    return wrapper

def async_chat(gen_ai_endpoint, version, environment, application_name,
                 tracer, pricing_info, capture_message_content, metrics, disable_metrics):
    """
    Creates a wrapper around a function call to trace and log its execution metrics.

    This function wraps any given function to measure its execution time,
    log its operation, and trace its execution using OpenTelemetry.
    
    Parameters:
    - version (str): The version of the Langchain application.
    - environment (str): The deployment environment (e.g., 'production', 'development').
    - application_name (str): Name of the Langchain application.
    - tracer (opentelemetry.trace.Tracer): The tracer object used for OpenTelemetry tracing.
    - pricing_info (dict): Information about the pricing for internal metrics (currently not used).
    - capture_message_content (bool): Flag indicating whether to trace the content of the response.

    Returns:
    - function: A higher-order function that takes a function 'wrapped' and returns
                a new function that wraps 'wrapped' with additional tracing and logging.
    """

    async def wrapper(wrapped, instance, args, kwargs):
        """
        An inner wrapper function that executes the wrapped function, measures execution
        time, and records trace data using OpenTelemetry.

        Parameters:
        - wrapped (Callable): The original function that this wrapper will execute.
        - instance (object): The instance to which the wrapped function belongs. This
                             is used for instance methods. For static and classmethods,
                             this may be None.
        - args (tuple): Positional arguments passed to the wrapped function.
        - kwargs (dict): Keyword arguments passed to the wrapped function.

        Returns:
        - The result of the wrapped function call.
        
        The wrapper initiates a span with the provided tracer, sets various attributes
        on the span based on the function's execution and response, and ensures
        errors are handled and logged appropriately.
        """

        server_address, server_port = "NOT_FOUND", "NOT_FOUND"

        if hasattr(instance, "model_id"):
            request_model = instance.model_id
        elif hasattr(instance, "model"):
            request_model = instance.model
        elif hasattr(instance, "model_name"):
            request_model = instance.model_name
        else:
            request_model = "NOT_FOUND"

        span_name = f"{SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT} {request_model}"

        with tracer.start_as_current_span(span_name, kind=SpanKind.CLIENT) as span:
            start_time = time.time()
            response = await wrapped(*args, **kwargs)
            end_time = time.time()

            try:
                # Format 'messages' into a single string
                message_prompt = kwargs.get("messages", "") or args[0]
                formatted_messages = []

                for message in message_prompt:
                    # Handle the case where message is a tuple
                    if isinstance(message, tuple) and len(message) == 2:
                        role, content = message
                    # Handle the case where message is a dictionary
                    elif isinstance(message, dict):
                        role = message["role"]
                        content = message["content"]
                    else:
                        continue

                    # Check if the content is a list
                    if isinstance(content, list):
                        content_str = ", ".join(
                            f'{item["type"]}: {item["text"] if "text" in item else item["image_url"]}'
                            if "type" in item else f'text: {item["text"]}'
                            for item in content
                        )
                        formatted_messages.append(f"{role}: {content_str}")
                    else:
                        formatted_messages.append(f"{role}: {content}")

                # Join all formatted messages with newline
                prompt = "\n".join(formatted_messages)

                input_tokens = general_tokens(str(prompt))
                output_tokens = general_tokens(str(response))

                # Calculate cost of the operation
                cost = get_chat_model_cost(
                    request_model,
                    pricing_info, input_tokens, output_tokens
                )

                try:
                    llm_response = response.content
                except AttributeError:
                    llm_response = response

                # Set base span attribues (OTel Semconv)
                span.set_attribute(TELEMETRY_SDK_NAME, "openlit")
                span.set_attribute(SemanticConvention.GEN_AI_OPERATION,
                                    SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT)
                span.set_attribute(SemanticConvention.GEN_AI_SYSTEM,
                                    SemanticConvention.GEN_AI_SYSTEM_LANGCHAIN)
                span.set_attribute(SemanticConvention.GEN_AI_REQUEST_MODEL,
                                    request_model)
                span.set_attribute(SemanticConvention.GEN_AI_RESPONSE_MODEL,
                                    request_model)
                span.set_attribute(SemanticConvention.GEN_AI_REQUEST_TEMPERATURE,
                                    str(getattr(instance, 'temperature', 1)))
                span.set_attribute(SemanticConvention.GEN_AI_REQUEST_TOP_K,
                                    str(getattr(instance, 'top_k', 1)))
                span.set_attribute(SemanticConvention.GEN_AI_REQUEST_TOP_P,
                                    str(getattr(instance, 'top_p', 1)))
                span.set_attribute(SemanticConvention.GEN_AI_USAGE_INPUT_TOKENS,
                                    input_tokens)
                span.set_attribute(SemanticConvention.GEN_AI_USAGE_OUTPUT_TOKENS,
                                    output_tokens)
                span.set_attribute(SemanticConvention.SERVER_ADDRESS,
                                    server_address)
                span.set_attribute(SemanticConvention.SERVER_PORT,
                                    server_port)

                # Set base span attribues (Extras)
                span.set_attribute(DEPLOYMENT_ENVIRONMENT,
                                     environment)
                span.set_attribute(SERVICE_NAME,
                                    application_name)
                span.set_attribute(SemanticConvention.GEN_AI_REQUEST_IS_STREAM,
                                    False)
                span.set_attribute(SemanticConvention.GEN_AI_USAGE_TOTAL_TOKENS,
                                    input_tokens + output_tokens)
                span.set_attribute(SemanticConvention.GEN_AI_USAGE_COST,
                                    cost)
                span.set_attribute(SemanticConvention.GEN_AI_SERVER_TTFT,
                                    end_time - start_time)
                span.set_attribute(SemanticConvention.GEN_AI_SDK_VERSION,
                                    version)

                if capture_message_content:
                    span.add_event(
                        name=SemanticConvention.GEN_AI_CONTENT_PROMPT_EVENT,
                        attributes={
                            SemanticConvention.GEN_AI_CONTENT_PROMPT: prompt,
                        },
                    )
                    span.add_event(
                        name=SemanticConvention.GEN_AI_CONTENT_COMPLETION_EVENT,
                        attributes={
                            SemanticConvention.GEN_AI_CONTENT_COMPLETION: llm_response,
                        },
                    )

                span.set_status(Status(StatusCode.OK))

                if disable_metrics is False:
                    attributes = create_metrics_attributes(
                        service_name=application_name,
                        deployment_environment=environment,
                        operation=SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT,
                        system=SemanticConvention.GEN_AI_SYSTEM_LANGCHAIN,
                        request_model=request_model,
                        server_address=server_address,
                        server_port=server_port,
                        response_model=request_model,
                    )

                    metrics["genai_client_usage_tokens"].record(
                        input_tokens + output_tokens, attributes
                    )
                    metrics["genai_client_operation_duration"].record(
                        end_time - start_time, attributes
                    )
                    metrics["genai_server_ttft"].record(
                        end_time - start_time, attributes
                    )
                    metrics["genai_requests"].add(1, attributes)
                    metrics["genai_completion_tokens"].add(output_tokens, attributes)
                    metrics["genai_prompt_tokens"].add(input_tokens, attributes)
                    metrics["genai_cost"].record(cost, attributes)

                # Return original response
                return response

            except Exception as e:
                handle_exception(span, e)
                logger.error("Error in trace creation: %s", e)

                # Return original response
                return response

    return wrapper



"""
Initializer of Auto Instrumentation of HuggingFace Transformer Functions
"""

from typing import Collection
import importlib.metadata
from opentelemetry.instrumentation.instrumentor import BaseInstrumentor
from wrapt import wrap_function_wrapper

from openlit.instrumentation.transformers.transformers import pipeline_wrapper

_instruments = ("transformers >= 4.48.0",)

class TransformersInstrumentor(BaseInstrumentor):
    """
    An instrumentor for HuggingFace Transformer library.
    """

    def instrumentation_dependencies(self) -> Collection[str]:
        return _instruments

    def _instrument(self, **kwargs):
        application_name = kwargs.get("application_name")
        environment = kwargs.get("environment")
        tracer = kwargs.get("tracer")
        metrics = kwargs.get("metrics_dict")
        pricing_info = kwargs.get("pricing_info")
        capture_message_content = kwargs.get("capture_message_content")
        disable_metrics = kwargs.get("disable_metrics")
        version = importlib.metadata.version("transformers")

        wrap_function_wrapper(
            "transformers",
            "TextGenerationPipeline.__call__",
            pipeline_wrapper(
                version,
                environment,
                application_name,
                tracer,
                pricing_info,
                capture_message_content,
                metrics,
                disable_metrics
            ),
        )

    def _uninstrument(self, **kwargs):
        pass



"""
Anthropic OpenTelemetry instrumentation utility functions
"""
import time

from opentelemetry.sdk.resources import SERVICE_NAME, TELEMETRY_SDK_NAME, DEPLOYMENT_ENVIRONMENT
from opentelemetry.trace import Status, StatusCode

from openlit.__helpers import (
    calculate_ttft,
    response_as_dict,
    calculate_tbt,
    extract_and_format_input,
    get_chat_model_cost,
    create_metrics_attributes,
    otel_event,
    concatenate_all_contents
)
from openlit.semcov import SemanticConvention

def process_chunk(self, chunk):
    """
    Process a chunk of response data and update state.
    """

    end_time = time.time()
    # Record the timestamp for the current chunk
    self._timestamps.append(end_time)

    if len(self._timestamps) == 1:
        # Calculate time to first chunk
        self._ttft = calculate_ttft(self._timestamps, self._start_time)

    chunked = response_as_dict(chunk)

    # Collect message IDs and input token from events
    if chunked.get('type') == 'message_start':
        self._response_id = chunked.get('message').get('id')
        self._input_tokens = chunked.get('message').get('usage').get('input_tokens')
        self._response_model = chunked.get('message').get('model')
        self._response_role = chunked.get('message').get('role')

    # Collect message IDs and aggregated response from events
    if chunked.get('type') == 'content_block_delta':
        if chunked.get('delta').get('text'):
            self._llmresponse += chunked.get('delta').get('text')
        elif chunked.get('delta').get('partial_json'):
            self._tool_arguments += chunked.get('delta').get('partial_json')

    if chunked.get('type') == 'content_block_start':
        if chunked.get('content_block').get('id'):
            self._tool_id = chunked.get('content_block').get('id')
        if chunked.get('content_block').get('name'):
            self._tool_name = chunked.get('content_block').get('name')

    # Collect output tokens and stop reason from events
    if chunked.get('type') == 'message_delta':
        self._output_tokens = chunked.get('usage').get('output_tokens')
        self._finish_reason = chunked.get('delta').get('stop_reason')

def common_chat_logic(scope, pricing_info, environment, application_name, metrics,
                        event_provider, capture_message_content, disable_metrics, version, is_stream):
    """
    Process chat request and generate Telemetry
    """

    scope._end_time = time.time()
    if len(scope._timestamps) > 1:
        scope._tbt = calculate_tbt(scope._timestamps)

    formatted_messages = extract_and_format_input(scope._kwargs.get('messages', ''))
    request_model = scope._kwargs.get('model', 'claude-3-opus-20240229')

    cost = get_chat_model_cost(request_model, pricing_info, scope._input_tokens, scope._output_tokens)

    # Set Span attributes (OTel Semconv)
    scope._span.set_attribute(TELEMETRY_SDK_NAME, 'openlit')
    scope._span.set_attribute(SemanticConvention.GEN_AI_OPERATION, SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT)
    scope._span.set_attribute(SemanticConvention.GEN_AI_SYSTEM, SemanticConvention.GEN_AI_SYSTEM_ANTHROPIC)
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_MODEL, request_model)
    scope._span.set_attribute(SemanticConvention.SERVER_PORT, scope._server_port)
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_MAX_TOKENS, scope._kwargs.get('max_tokens', -1))
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_STOP_SEQUENCES, scope._kwargs.get('stop_sequences', []))
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_TEMPERATURE, scope._kwargs.get('temperature', 1.0))
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_TOP_K, scope._kwargs.get('top_k', 1.0))
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_TOP_P, scope._kwargs.get('top_p', 1.0))
    scope._span.set_attribute(SemanticConvention.GEN_AI_RESPONSE_FINISH_REASON, [scope._finish_reason])
    scope._span.set_attribute(SemanticConvention.GEN_AI_RESPONSE_ID, scope._response_id)
    scope._span.set_attribute(SemanticConvention.GEN_AI_RESPONSE_MODEL, scope._response_model)
    scope._span.set_attribute(SemanticConvention.GEN_AI_USAGE_INPUT_TOKENS, scope._input_tokens)
    scope._span.set_attribute(SemanticConvention.GEN_AI_USAGE_OUTPUT_TOKENS, scope._output_tokens)
    scope._span.set_attribute(SemanticConvention.SERVER_ADDRESS, scope._server_address)

    scope._span.set_attribute(SemanticConvention.GEN_AI_OUTPUT_TYPE,
                              'text' if isinstance(scope._llmresponse, str) else 'json')

    scope._span.set_attribute(DEPLOYMENT_ENVIRONMENT, environment)
    scope._span.set_attribute(SERVICE_NAME, application_name)
    scope._span.set_attribute(SemanticConvention.GEN_AI_REQUEST_IS_STREAM, is_stream)
    scope._span.set_attribute(SemanticConvention.GEN_AI_CLIENT_TOKEN_USAGE, scope._input_tokens + scope._output_tokens)
    scope._span.set_attribute(SemanticConvention.GEN_AI_USAGE_COST, cost)
    scope._span.set_attribute(SemanticConvention.GEN_AI_SERVER_TBT, scope._tbt)
    scope._span.set_attribute(SemanticConvention.GEN_AI_SERVER_TTFT, scope._ttft)
    scope._span.set_attribute(SemanticConvention.GEN_AI_SDK_VERSION, version)

    # To be removed one the change to log events (from span events) is complete
    prompt = concatenate_all_contents(formatted_messages)
    if capture_message_content:
        scope._span.add_event(
            name=SemanticConvention.GEN_AI_CONTENT_PROMPT_EVENT,
            attributes={
                SemanticConvention.GEN_AI_CONTENT_PROMPT: prompt,
            },
        )
        scope._span.add_event(
            name=SemanticConvention.GEN_AI_CONTENT_COMPLETION_EVENT,
            attributes={
                SemanticConvention.GEN_AI_CONTENT_COMPLETION: scope._llmresponse,
            },
        )

    choice_event_body = {
        'finish_reason': scope._finish_reason,
        'index': 0,
        'message': {
            **({'content': scope._llmresponse} if capture_message_content else {}),
            'role': scope._response_role
        }
    }

    if scope._tool_calls:
        choice_event_body['message'].update({
            'tool_calls': {
                'function': {
                    'name': scope._tool_calls.get('name', ''),
                    'arguments': scope._tool_calls.get('input', '')
                },
                'id': scope._tool_calls.get('id', ''),
                'type': 'function'
            }
        })

    # Emit events
    for role in ['user', 'system', 'assistant', 'tool']:
        if formatted_messages.get(role, {}).get('content', ''):
            event = otel_event(
                name=getattr(SemanticConvention, f'GEN_AI_{role.upper()}_MESSAGE'),
                attributes={
                    SemanticConvention.GEN_AI_SYSTEM: SemanticConvention.GEN_AI_SYSTEM_ANTHROPIC
                },
                body = {
                    # pylint: disable=line-too-long
                    **({'content': formatted_messages.get(role, {}).get('content', '')} if capture_message_content else {}),
                    'role': formatted_messages.get(role, {}).get('role', []),
                    **({
                        'tool_calls': {
                            'function': {
                                # pylint: disable=line-too-long
                                'name': (scope._tool_calls[0].get('function', {}).get('name', '') if scope._tool_calls else ''),
                                'arguments': (scope._tool_calls[0].get('function', {}).get('arguments', '') if scope._tool_calls else '')
                            },
                            'id': (scope._tool_calls[0].get('id', '') if scope._tool_calls else ''),
                            'type': 'function'
                        }
                    } if role == 'assistant' else {}),
                    **({
                        'id': (scope._tool_calls[0].get('id', '') if scope._tool_calls else '')
                    } if role == 'tool' else {})
                }
            )
            event_provider.emit(event)

    choice_event = otel_event(
        name=SemanticConvention.GEN_AI_CHOICE,
        attributes={
            SemanticConvention.GEN_AI_SYSTEM: SemanticConvention.GEN_AI_SYSTEM_ANTHROPIC
        },
        body=choice_event_body
    )
    event_provider.emit(choice_event)

    scope._span.set_status(Status(StatusCode.OK))

    if not disable_metrics:
        metrics_attributes = create_metrics_attributes(
            service_name=application_name,
            deployment_environment=environment,
            operation=SemanticConvention.GEN_AI_OPERATION_TYPE_CHAT,
            system=SemanticConvention.GEN_AI_SYSTEM_ANTHROPIC,
            request_model=request_model,
            server_address=scope._server_address,
            server_port=scope._server_port,
            response_model=scope._response_model,
        )

        metrics['genai_client_usage_tokens'].record(scope._input_tokens + scope._output_tokens, metrics_attributes)
        metrics['genai_client_operation_duration'].record(scope._end_time - scope._start_time, metrics_attributes)
        metrics['genai_server_tbt'].record(scope._tbt, metrics_attributes)
        metrics['genai_server_ttft'].record(scope._ttft, metrics_attributes)
        metrics['genai_requests'].add(1, metrics_attributes)
        metrics['genai_completion_tokens'].add(scope._output_tokens, metrics_attributes)
        metrics['genai_prompt_tokens'].add(scope._input_tokens, metrics_attributes)
        metrics['genai_cost'].record(cost, metrics_attributes)

def process_streaming_chat_response(self, pricing_info, environment, application_name, metrics,
                                    event_provider, capture_message_content=False, disable_metrics=False, version=''):
    """
    Process chat request and generate Telemetry
    """
    if self._tool_id != '':
        self._tool_calls = {
            'id': self._tool_id,
            'name': self._tool_name,
            'input': self._tool_arguments
        }

    common_chat_logic(self, pricing_info, environment, application_name, metrics,
                        event_provider, capture_message_content, disable_metrics, version, is_stream=True)

def process_chat_response(response, request_model, pricing_info, server_port, server_address,
                          environment, application_name, metrics, event_provider, start_time,
                          span, capture_message_content=False, disable_metrics=False, version='1.0.0', **kwargs):
    """
    Process chat request and generate Telemetry
    """

    self = type('GenericScope', (), {})()
    response_dict = response_as_dict(response)

    # pylint: disable = no-member
    self._start_time = start_time
    self._end_time = time.time()
    self._span = span
    self._llmresponse = response_dict.get('content', {})[0].get('text', '')
    self._response_role = response_dict.get('message', {}).get('role', 'assistant')
    self._input_tokens = response_dict.get('usage').get('input_tokens')
    self._output_tokens = response_dict.get('usage').get('output_tokens')
    self._response_model = response_dict.get('model', '')
    self._finish_reason = response_dict.get('stop_reason', '')
    self._response_id = response_dict.get('id', '')
    self._timestamps = []
    self._ttft, self._tbt = self._end_time - self._start_time, 0
    self._server_address, self._server_port = server_address, server_port
    self._kwargs = kwargs
    #pylint: disable=line-too-long
    self._tool_calls = (lambda c: c[1] if len(c) > 1 and c[1].get('type') == 'tool_use' else None)(response_dict.get('content', []))

    common_chat_logic(self, pricing_info, environment, application_name, metrics,
                        event_provider, capture_message_content, disable_metrics, version, is_stream=False)

    return response



"""
Module for monitoring AG2 API calls.
"""

import logging
from opentelemetry.trace import SpanKind, Status, StatusCode
from opentelemetry.sdk.resources import SERVICE_NAME, TELEMETRY_SDK_NAME, DEPLOYMENT_ENVIRONMENT
from openlit.__helpers import (
    handle_exception,
)
from openlit.semcov import SemanticConvention

# Initialize logger for logging potential issues and operations
logger = logging.getLogger(__name__)

def set_span_attributes(span, version, operation_name, environment,
        application_name, server_address, server_port, request_model):
    """
    Set common attributes for the span.
    """

    # Set Span attributes (OTel Semconv)
    span.set_attribute(TELEMETRY_SDK_NAME, 'openlit')
    span.set_attribute(SemanticConvention.GEN_AI_OPERATION, operation_name)
    span.set_attribute(SemanticConvention.GEN_AI_SYSTEM, SemanticConvention.GEN_AI_SYSTEM_AG2)
    span.set_attribute(SemanticConvention.SERVER_ADDRESS, server_address)
    span.set_attribute(SemanticConvention.SERVER_PORT, server_port)
    span.set_attribute(SemanticConvention.GEN_AI_REQUEST_MODEL, request_model)

    # Set Span attributes (Extras)
    span.set_attribute(DEPLOYMENT_ENVIRONMENT, environment)
    span.set_attribute(SERVICE_NAME, application_name)
    span.set_attribute(SemanticConvention.GEN_AI_SDK_VERSION, version)

def create_agent(version, environment, application_name,
                      tracer, event_provider, pricing_info, capture_message_content, metrics, disable_metrics):
    """
    Generates a telemetry wrapper for GenAI function call
    """
    def wrapper(wrapped, instance, args, kwargs):
        server_address, server_port = '127.0.0.1', 80

        agent_name = kwargs.get('name', 'openai_agent')
        span_name = f'{SemanticConvention.GEN_AI_OPERATION_TYPE_CREATE_AGENT} {agent_name}'

        with tracer.start_as_current_span(span_name, kind=SpanKind.CLIENT) as span:
            try:
                response = wrapped(*args, **kwargs)

                set_span_attributes(span, version, SemanticConvention.GEN_AI_OPERATION_TYPE_CREATE_AGENT,
                    environment, application_name, server_address, server_port, kwargs.get('model', 'gpt-4o'))
                span.set_attribute(SemanticConvention.GEN_AI_AGENT_NAME, agent_name)

                span.set_attribute(SemanticConvention.GEN_AI_AGENT_DESCRIPTION, kwargs.get('instructions', ''))

                span.set_status(Status(StatusCode.OK))

                return response

            except Exception as e:
                handle_exception(span, e)
                logger.error('Error in trace creation: %s', e)
                return response

    return wrapper
